"""
Utility functions.
"""
import asyncio
import collections
import functools
import inspect
import io
import logging
import os
from typing import Set  # noqa

import libnacl
import logbook
import logbook.compat
import logbook.more
import wrapt
from memoization.caching.general import (
    keys_order_dependent,
    values_with_ttl,
)

from .key import Key

__all__ = (
    'enable_logging',
    'disable_logging',
    'get_logger',
    'read_key_or_key_file',
    'raise_server_error',
    'randint',
    'ViewIOReader',
    'ViewIOWriter',
    'async_ttl_cache',
    'aio_run',
    'aio_run_proxy',
    'AioRunMixin',
)

_logger_group = logbook.LoggerGroup()
_logger_group.disabled = True
_logger_redirect_handler = logbook.compat.RedirectLoggingHandler()
_logger_convert_level_handler = logbook.compat.LoggingHandler()


def _convert_level(logging_level):
    return _logger_convert_level_handler.convert_level(logging_level)


def enable_logging(level=logbook.WARNING, asyncio_level=None, aiohttp_level=None):
    # Determine levels
    level = logbook.lookup_level(level)
    converted_level = _convert_level(level)
    if asyncio_level is None:
        asyncio_level = converted_level
    else:
        asyncio_level = _convert_level(asyncio_level)
    if aiohttp_level is None:
        aiohttp_level = converted_level
    else:
        aiohttp_level = _convert_level(aiohttp_level)

    # Enable logger group
    _logger_group.disabled = False

    # Enable asyncio debug logging
    os.environ['PYTHONASYNCIODEBUG'] = '1'

    # Redirect asyncio logger
    logger = logging.getLogger('asyncio')
    logger.setLevel(asyncio_level)
    logger.addHandler(_logger_redirect_handler)

    # Redirect aiohttp logger
    logger = logging.getLogger('aiohttp')
    logger.setLevel(aiohttp_level)
    logger.addHandler(_logger_redirect_handler)


def disable_logging():
    # Reset aiohttp logger
    logger = logging.getLogger('aiohttp')
    logger.removeHandler(_logger_redirect_handler)
    logger.setLevel(logging.NOTSET)

    # Reset asyncio logger
    logger = logging.getLogger('asyncio')
    logger.removeHandler(_logger_redirect_handler)
    logger.setLevel(logging.NOTSET)

    # Disable asyncio debug logging
    del os.environ['PYTHONASYNCIODEBUG']

    # Disable logger group
    _logger_group.disabled = True


def get_logger(name=None, level=logbook.NOTSET):
    """
    Return a :class:`logbook.Logger`.

    Arguments:
        - `name`: The name of a specific sub-logger.
    """
    base_name = 'threema.gateway'
    name = base_name if name is None else '.'.join((base_name, name))

    # Create new logger and add to group
    logger = logbook.Logger(name=name, level=level)
    _logger_group.add_logger(logger)
    return logger


# TODO: Raises
def read_key_or_key_file(key, expected_type):
    """
    Decode a hex-encoded key or read it from a file.

    Arguments:
        - `key`: A hex-encoded key or the name of a file which contains
          a key.
        - `expected_type`: One of the types of :class:`Key.Type`.

    Return a:class:`libnacl.public.SecretKey` or
    :class:`libnacl.public.PublicKey` instance.
    """
    # Read key file (if any)
    try:
        with open(key) as file:
            key = file.readline().strip()
    except IOError:
        pass

    # Convert to key instance
    return Key.decode(key, expected_type)


async def raise_server_error(response, error):
    """
    Raise a :class:`GatewayServerError` exception from a
    HTTP response. Releases the response before raising.


    Arguments:
        - `response`: A :class:`aiohttp.ClientResponse` instance.
        - `error`: The :class:`GatewayServerError`. to instantiate.

    Always raises :class:`GatewayServerError`.
    """
    status = response.status
    await response.release()
    raise error(status)


def randint(a, b):
    """
    Return a cryptographically secure random integer N such that
    ``a <= N <= b``.
    """
    n = libnacl.randombytes_uniform(b) + a
    assert a <= n <= b
    return n


# TODO: Document properly
class ViewIOReader(io.RawIOBase):
    def __init__(self, bytes_or_view):
        super().__init__()
        if isinstance(bytes_or_view, bytes):
            bytes_or_view = memoryview(bytes_or_view)
        self._view = bytes_or_view
        self._offset = 0
        self._length = len(self._view)

    # IOBase methods

    def fileno(self):
        raise OSError('No file descriptors used')

    def isatty(self):
        return False

    def readable(self):
        return True

    def readline(self, size=-1):
        raise NotImplementedError

    def readlines(self, hint=-1):
        raise NotImplementedError

    def seek(self, offset, whence=os.SEEK_SET):
        if whence == os.SEEK_SET:
            pass
        elif whence == os.SEEK_CUR:
            offset += self._offset
        elif whence == os.SEEK_END:
            offset = self._length - offset
        else:
            raise ValueError('Invalid whence value')
        if not 0 < offset <= self._length:
            raise ValueError('Offset is greater than view length')
        self._offset = offset
        return offset

    def seekable(self):
        return True

    def tell(self):
        return self._offset

    def writable(self):
        return False

    # RawIOBase methods

    def read(self, size=-1):
        if size == -1:
            return self.readall()
        elif size < 0:
            raise ValueError('Negative size')
        start, end = self._offset, min(self._offset + size, self._length)
        self._offset = end
        return self._view[start:end]

    def readall(self):
        return self.read(self._length - self._offset)

    def readinto(self, b):
        data = self.readall()
        b.extend(data)
        return len(data)

    # Custom methods

    def __len__(self):
        return self._length - self._offset

    def readexactly(self, size):
        data = self.read(size)
        if len(data) < size:
            raise asyncio.IncompleteReadError(data, size)
        else:
            return data


# TODO: Document properly
class ViewIOWriter(io.RawIOBase):
    def __init__(self, bytes_or_views=None):
        super().__init__()
        self._views = []
        self._length = 0
        if bytes_or_views is not None:
            for bytes_or_view in bytes_or_views:
                self.writeexactly(bytes_or_view)

    # IOBase methods

    def fileno(self):
        raise OSError('No file descriptors used')

    def isatty(self):
        return False

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return True

    # RawIOBase methods

    def write(self, bytes_or_view):
        # Convert to memoryview if necessary
        if isinstance(bytes_or_view, bytes):
            bytes_or_view = memoryview(bytes_or_view)

        # Append
        length = len(bytes_or_view)
        self._length += length
        self._views.append(bytes_or_view)
        return length

    def writelines(self, lines):
        raise NotImplementedError

    # Custom methods

    def __radd__(self, other):
        self.extend(other)
        return self

    def __len__(self):
        return self._length

    def getvalue(self):
        return b''.join(self._views)

    # noinspection PyProtectedMember
    def extend(self, other):
        self._views += other._views
        self._length += other._length

    def writeexactly(self, bytes_or_view):
        return self.write(bytes_or_view)


_CacheInfo = collections.namedtuple('_CacheInfo', ('hits', 'misses', 'length', 'ttl'))


def async_ttl_cache(ttl):
    """
    Cache with expiration (TTL) for asyncio coroutines.

    If *ttl* is set, cached values will be cleared after
    *ttl* seconds.

    View the cache statistics named tuple (hits, misses, maxsize,
    currsize) with f.cache_info().  Clear the cache and statistics
    with f.cache_clear(). Access the underlying function with
    f.__wrapped__.
    """
    def _decorator(func):
        # Ensure it is a coroutine
        if not asyncio.iscoroutinefunction(func):
            raise ValueError('Function is not a coroutine')

        # Cache storage
        cache = {}
        hits = misses = 0

        async def _wrapper(*args, **kwargs):
            nonlocal hits, misses
            key = keys_order_dependent.make_key(args, kwargs)

            # Attempt to get value from cache
            try:
                node = cache[key]
            except KeyError:
                pass
            else:
                # Hit. Return cached value if still valid.
                if values_with_ttl.is_cache_value_valid(node):
                    hits += 1
                    return values_with_ttl.retrieve_result_from_cache_value(node)

            # Miss. Get value from function and insert into cache.
            misses += 1
            value = await func(*args, **kwargs)
            cache[key] = values_with_ttl.make_cache_value(value, ttl)
            return value

        def cache_clear():
            """
            Clear the cache and its statistics information
            """
            nonlocal hits, misses
            cache.clear()
            hits = misses = 0

        def cache_info():
            """
            Show statistics information
            """
            return _CacheInfo(hits, misses, len(cache), ttl)

        # Expose functions to wrapper
        _wrapper.cache_clear = cache_clear
        _wrapper.cache_info = cache_info
        _wrapper._cache = cache

        return functools.update_wrapper(_wrapper, func)
    return _decorator


def aio_run(func):
    """
    Decorate an async function to run as a normal blocking function.

    The async function will be executed in the currently running event
    loop (or automatically create one if none exists).

    Example:

    .. code-block::
        async def coroutine(timeout):
            await asyncio.sleep(timeout)
            return True

        @aio_run
        async def runner(*args, **kwargs):
            return await coroutine(*args, **kwargs)

        # Call coroutine in a blocking manner
        result = runner(timeout=1.0)
        print(result)
    """
    def _wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(func(*args, **kwargs))
    return functools.update_wrapper(_wrapper, func)


def aio_run_proxy(cls):
    """
    Proxy a publicly accessible class and run all methods marked as
    async inside it (using the class attribute `async_functions`) with
    an event loop to make it appear as a traditional blocking method.

    Arguments:
        - `cls`: A class to be wrapped. The class must inherit
          :class:`AioRunMixin`. The class and all base classes must
          supply a class attribute `async_functions` which is an
          iterable of method names that should appear as traditional
          blocking functions from the outside.

    Returns a class factory.

    .. note:: The `unwrap` property of the resulting instance can be
              used to get the original instance.
    """
    # Ensure each base class has added a class-level iterable of async functions
    async_functions = set()
    for base_class in inspect.getmro(cls)[:-1]:
        try:
            async_functions.update(base_class.__dict__.get('async_functions', None))
        except TypeError:
            message = "Class {} is missing 'async_functions' iterable"
            raise ValueError(message.format(base_class.__name__))

    # Sanity-check
    if not issubclass(cls, AioRunMixin):
        raise TypeError("Class {} did not inherit 'AioRunMixin'".format(
            cls.__name__))

    class _AioRunProxyDecoratorFactory(wrapt.ObjectProxy):
        def __call__(self, *args, **kwargs):
            # Create the instance while an event loop is running
            loop = asyncio.get_event_loop()
            if loop.is_running():
                instance = cls(*args, **kwargs)
            else:
                async def _create_instance():
                    return cls(*args, **kwargs)
                instance = loop.run_until_complete(_create_instance())

            # Sanity-check
            if not isinstance(instance, AioRunMixin):
                raise TypeError("Class {} did not inherit 'AioRunMixin'".format(
                    cls.__name__))

            # Wrap with proxy (if required)
            if instance.blocking:
                class _AioRunProxy(wrapt.ObjectProxy):
                    @property
                    def unwrap(self):
                        """
                        Get the wrapped instance.
                        """
                        return self.__wrapped__

                # Wrap all async functions with `aio_run`
                for name in async_functions:
                    def _method(instance_, name_, *args_, **kwargs_):
                        method = aio_run(getattr(instance_, name_))
                        return method(*args_, **kwargs_)

                    _method = functools.partial(_method, instance, name)
                    setattr(_AioRunProxy, name, _method)

                return _AioRunProxy(instance)
            else:
                return instance

    return _AioRunProxyDecoratorFactory(cls)


class AioRunMixin:
    """
    Must be inherited when using :func:`aio_run_proxy`.

    Arguments:
        - `blocking`: Switch to turn the blocking API on or off.
    """
    async_functions = set()  # type: Set[str]

    def __init__(self, blocking=False):
        self.blocking = blocking

    @property
    def unwrap(self):
        """
        Get the wrapped instance.
        """
        return self



import asyncio
import asyncio.subprocess
import binascii
import collections
import copy
import hashlib
import hmac
import os
import socket
import ssl
import subprocess
import sys
import time
from contextlib import closing

import aiohttp
import pytest
from aiohttp import web

import threema.gateway
from threema.gateway import e2e
from threema.gateway.key import Key

_res_path = os.path.normpath(os.path.join(
    os.path.abspath(__file__), os.pardir, 'res'))


class RawMessage(e2e.Message):
    def __init__(self, connection, nonce=None, message=None, **kwargs):
        super().__init__(connection, e2e.Message.Type.text_message, **kwargs)
        self.nonce = nonce
        self.message = message

    async def pack(self, writer):
        raise NotImplementedError

    @classmethod
    async def unpack(cls, connection, parameters, key_pair, reader):
        raise NotImplementedError

    async def send(self, get_data_only=False):
        """
        Send the raw message

        Return the ID of the message.
        """
        # Send message
        if get_data_only:
            return self.nonce, self.message
        else:
            return await self._connection.send_e2e(**{
                'to': self.to_id,
                'nonce': binascii.hexlify(self.nonce).decode(),
                'box': binascii.hexlify(self.message).decode()
            })


class Server:
    def __init__(self):
        self.threema_jpg = os.path.join(_res_path, 'threema.jpg')
        self.threema_mp4 = os.path.join(_res_path, 'threema.mp4')
        key = b'4a6a1b34dcef15d43cb74de2fd36091be99fbbaf126d099d47d83d919712c72b'
        self.echoecho_key = key
        self.echoecho_encoded_key = 'public:' + key.decode('ascii')
        decoded_private_key = Key.decode(pytest.msgapi.private, Key.Type.private)
        self.mocking_key = Key.derive_public(decoded_private_key).hex_pk()
        self.blobs = {}
        self.latest_blob_ids = []
        self.routes = [
            web.get('/pubkeys/{key}', self.pubkeys),
            web.get('/lookup/phone/{phone}', self.lookup_phone),
            web.get('/lookup/phone_hash/{phone_hash}', self.lookup_phone_hash),
            web.get('/lookup/email/{email}', self.lookup_email),
            web.get('/lookup/email_hash/{email_hash}', self.lookup_email_hash),
            web.get('/capabilities/{id}', self.capabilities),
            web.get('/credits', self.credits),
            web.post('/send_simple', self.send_simple),
            web.post('/send_e2e', self.send_e2e),
            web.post('/upload_blob', self.upload_blob),
            web.get('/blobs/{blob_id}', self.download_blob),
        ]

    async def pubkeys(self, request):
        key = request.match_info['key']
        from_, secret = request.query['from'], request.query['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif len(key) != 8:
            return web.Response(status=404)
        elif key == 'ECHOECHO':
            return web.Response(body=self.echoecho_key)
        elif key == '*MOCKING':
            return web.Response(body=self.mocking_key)
        return web.Response(status=404)

    async def lookup_phone(self, request):
        phone = request.match_info['phone']
        from_, secret = request.query['from'], request.query['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif not phone.isdigit():
            return web.Response(status=404)
        elif phone == '44123456789':
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    async def lookup_phone_hash(self, request):
        phone_hash = request.match_info['phone_hash']
        from_, secret = request.query['from'], request.query['secret']
        hash_ = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214c'
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif len(phone_hash) % 2 != 0:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=500)
        elif len(phone_hash) != 64:
            return web.Response(status=400)
        elif phone_hash == hash_:
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    async def lookup_email(self, request):
        email = request.match_info['email']
        from_, secret = request.query['from'], request.query['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif email == 'echoecho@example.com':
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    async def lookup_email_hash(self, request):
        email_hash = request.match_info['email_hash']
        from_, secret = request.query['from'], request.query['secret']
        hash_ = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e2'
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif len(email_hash) % 2 != 0:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=500)
        elif len(email_hash) != 64:
            return web.Response(status=400)
        elif email_hash == hash_:
            return web.Response(body=b'ECHOECHO')
        return web.Response(status=404)

    async def capabilities(self, request):
        id_ = request.match_info['id']
        from_, secret = request.query['from'], request.query['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        elif id_ == 'ECHOECHO':
            return web.Response(body=b'text,image,video,file')
        elif id_ == '*MOCKING':
            return web.Response(body=b'text,image,video,file')
        return web.Response(status=404)

    async def credits(self, request):
        from_, secret = request.query['from'], request.query['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)
        return web.Response(body=b'100')

    async def send_simple(self, request):
        post = await request.post()

        # Check API identity
        if (post['from'], post['secret']) not in pytest.msgapi.api_identities:
            return web.Response(status=401)

        # Get ID from to, email or phone
        if 'to' in post:
            id_ = post['to']
        elif post.get('email', None) == 'echoecho@example.com':
            id_ = 'ECHOECHO'
        elif post.get('phone', None) == '44123456789':
            id_ = 'ECHOECHO'
        else:
            return web.Response(status=404)

        # Process
        text = post['text']
        if post['from'] == pytest.msgapi.nocredit_id:
            return web.Response(status=402)
        elif id_ != 'ECHOECHO':
            return web.Response(status=400)
        elif len(text) > 3500:
            return web.Response(status=413)
        return web.Response(body=b'0' * 16)

    async def send_e2e(self, request):
        post = await request.post()

        # Check API identity
        if (post['from'], post['secret']) not in pytest.msgapi.api_identities:
            return web.Response(status=401)

        # Get ID, nonce and box
        id_ = post['to']
        nonce, box = binascii.unhexlify(post['nonce']), binascii.unhexlify(post['box'])

        # Process
        if post['from'] == pytest.msgapi.nocredit_id:
            return web.Response(status=402)
        elif id_ != 'ECHOECHO':
            return web.Response(status=400)
        elif len(nonce) != 24:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=400)
        elif len(box) > 4000:
            return web.Response(status=413)
        return web.Response(body=b'1' * 16)

    async def upload_blob(self, request):
        try:
            data = await request.post()

            # Check API identity
            api_identity = (request.query['from'], request.query['secret'])
            if api_identity not in pytest.msgapi.api_identities:
                return web.Response(status=401)
        except KeyError:
            return web.Response(status=401)

        try:
            # Get blob
            blob = data['blob'].file.read()
        except KeyError:
            # Note: This status code might not be intended and may change in the future
            return web.Response(status=500)

        # Generate ID
        blob_id = hashlib.sha256(blob).hexdigest()[:32]

        # Process
        if request.query['from'] == pytest.msgapi.nocredit_id:
            return web.Response(status=402)
        elif len(blob) == 0:
            return web.Response(status=400)
        elif len(blob) > 20 * (2**20):
            return web.Response(status=413)

        # Store blob and return
        self.blobs[blob_id] = blob
        self.latest_blob_ids.append(blob_id)
        return web.Response(body=blob_id.encode())

    async def download_blob(self, request):
        blob_id = request.match_info['blob_id']

        # Check API identity
        from_, secret = request.query['from'], request.query['secret']
        if (from_, secret) not in pytest.msgapi.api_identities:
            return web.Response(status=401)

        # Get blob
        try:
            blob = self.blobs[blob_id]
        except KeyError:
            return web.Response(status=404)
        else:
            return web.Response(
                body=blob,
                content_type='application/octet-stream'
            )


def pytest_addoption(parser):
    help_ = 'loop: Use a different event loop, supported: asyncio, uvloop'
    parser.addoption("--loop", action="store", help=help_)


def pytest_report_header(config):
    return 'Using event loop: {}'.format(default_event_loop(config=config))


def pytest_namespace():
    private = 'private:dd9413d597092b004fedc4895db978425efa328ba1f1ec6729e46e09231b8a7e'
    public = Key.encode(Key.derive_public(Key.decode(private, Key.Type.private)))
    values = {'msgapi': {
        'cli_path': os.path.join(
            os.path.dirname(__file__),
            '../threema/gateway/bin/gateway_client.py',
        ),
        'cert_path': os.path.join(_res_path, 'cert.pem'),
        'base_url': 'https://msgapi.threema.ch',
        'ip': '127.0.0.1',
        'id': '*MOCKING',
        'secret': 'mock',
        'private': private,
        'public': public,
        'nocredit_id': 'NOCREDIT',
        'noexist_id': '*NOEXIST',
    }}
    values['msgapi']['api_identities'] = {
        (values['msgapi']['id'], values['msgapi']['secret']),
        (values['msgapi']['nocredit_id'], values['msgapi']['secret'])
    }
    return values


def default_event_loop(request=None, config=None):
    if request is not None:
        config = request.config
    loop = config.getoption("--loop")
    if loop == 'uvloop':
        import uvloop  # noqa
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    else:
        loop = 'asyncio'
    return loop


def unused_tcp_port():
    """
    Find an unused localhost TCP port from 1024-65535 and return it.
    """
    with closing(socket.socket()) as sock:
        sock.bind((pytest.msgapi.ip, 0))
        return sock.getsockname()[1]


def identity():
    return pytest.msgapi.id, pytest.msgapi.secret


@pytest.fixture(scope='module')
def server():
    return Server()


@pytest.fixture(scope='module')
def raw_message():
    return RawMessage


@pytest.fixture(scope='module')
def event_loop(request):
    """
    Create an instance of the requested event loop.
    """
    default_event_loop(request=request)

    # Close previous event loop
    policy = asyncio.get_event_loop_policy()
    policy.get_event_loop().close()

    # Create new event loop
    _event_loop = policy.new_event_loop()
    policy.set_event_loop(_event_loop)

    def fin():
        _event_loop.close()

    # Add finaliser and return new event loop
    request.addfinalizer(fin)
    return _event_loop


@pytest.fixture(scope='module')
def api_server_port():
    return unused_tcp_port()


@pytest.fixture(scope='module')
def api_server(request, event_loop, api_server_port, server):
    async def start_server():
        port = api_server_port
        app = web.Application(client_max_size=100 * (2**20))
        app.router.add_routes(server.routes)
        runner = web.AppRunner(app)
        await runner.setup()
        site = web.TCPSite(runner, host=pytest.msgapi.ip, port=port, shutdown_timeout=1.0)
        await site.start()
        return app, runner, site
    app, runner, site = event_loop.run_until_complete(start_server())

    def fin():
        async def stop_server():
            await site.stop()
            await runner.cleanup()
            await app.cleanup()
        event_loop.run_until_complete(stop_server())

    request.addfinalizer(fin)


@pytest.fixture(scope='module')
def mock_url(api_server_port):
    """
    Return the URL where the test server can be reached.
    """
    return 'http://{}:{}'.format(pytest.msgapi.ip, api_server_port)


@pytest.fixture(scope='module')
def connection(request, event_loop, api_server, mock_url):
    async def create_connection():
        # Note: We're not doing anything with the server but obviously the
        # server needs to be started to be able to connect
        return threema.gateway.Connection(
            identity=pytest.msgapi.id,
            secret=pytest.msgapi.secret,
            key=pytest.msgapi.private,
        )
    connection_ = event_loop.run_until_complete(create_connection())

    # Patch URLs
    connection_.urls = {key: value.replace(pytest.msgapi.base_url, mock_url)
                        for key, value in connection_.urls.items()}

    def fin():
        event_loop.run_until_complete(connection_.close())

    request.addfinalizer(fin)
    return connection_


@pytest.fixture(scope='module')
def connection_blocking(request, event_loop, api_server, mock_url):
    async def create_connection():
        # Note: We're not doing anything with the server but obviously the
        # server needs to be started to be able to connect
        return threema.gateway.Connection(
            identity=pytest.msgapi.id,
            secret=pytest.msgapi.secret,
            key=pytest.msgapi.private,
            blocking=True,
        )
    connection_ = event_loop.run_until_complete(create_connection())

    # Patch URLs
    connection_.urls = {key: value.replace(pytest.msgapi.base_url, mock_url)
                        for key, value in connection_.urls.items()}

    def fin():
        event_loop.run_until_complete(connection_.close())

    request.addfinalizer(fin)
    return connection_


@pytest.fixture(scope='module')
def invalid_connection(connection):
    invalid_connection_ = copy.copy(connection)
    invalid_connection_.id = pytest.msgapi.noexist_id
    return invalid_connection_


@pytest.fixture(scope='module')
def nocredit_connection(connection):
    nocredit_connection_ = copy.copy(connection)
    nocredit_connection_.id = pytest.msgapi.nocredit_id
    return nocredit_connection_


@pytest.fixture(scope='module')
def blob():
    return b'\x01\x02\x03'


@pytest.fixture(scope='module')
def blob_id(event_loop, connection, blob):
    coroutine = connection.upload(blob)
    return event_loop.run_until_complete(coroutine)


@pytest.fixture(scope='module')
def cli(api_server, api_server_port, event_loop):
    async def call_cli(*args, input=None, timeout=3.0):
        # Prepare environment
        env = os.environ.copy()
        env['THREEMA_TEST_API'] = str(api_server_port)
        test_api_mode = 'WARNING: Currently running in test mode!'

        # Call CLI in subprocess and get output
        parameters = [sys.executable, pytest.msgapi.cli_path] + list(args)
        if isinstance(input, str):
            input = input.encode('utf-8')

        # Create process
        create = asyncio.create_subprocess_exec(
            *parameters, env=env, stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT)
        process = await create

        # Wait for process to terminate
        coroutine = process.communicate(input=input)
        output, _ = await asyncio.wait_for(coroutine, timeout)

        # Process output
        output = output.decode('utf-8')
        if test_api_mode not in output:
            print(output)
            raise ValueError('Not running in test mode')

        # Strip leading empty lines and pydev debugger output
        rubbish = [
            'pydev debugger: process',
            'Traceback (most recent call last):',
            test_api_mode,
        ]
        lines = []
        skip_following_empty_lines = True
        for line in output.splitlines(keepends=True):
            if any((line.startswith(s) for s in rubbish)):
                skip_following_empty_lines = True
            elif not skip_following_empty_lines or len(line.strip()) > 0:
                lines.append(line)
                skip_following_empty_lines = False

        # Strip trailing empty lines
        empty_lines_count = 0
        for line in reversed(lines):
            if len(line.strip()) > 0:
                break
            empty_lines_count += 1
        if empty_lines_count > 0:
            lines = lines[:-empty_lines_count]
        output = ''.join(lines)

        # Check return code
        if process.returncode != 0:
            raise subprocess.CalledProcessError(process.returncode, parameters,
                                                output=output)
        return output
    return call_cli


@pytest.fixture(scope='module')
def private_key_file(tmpdir_factory):
    file = tmpdir_factory.mktemp('keys').join('private_key')
    file.write(pytest.msgapi.private)
    return str(file)


@pytest.fixture(scope='module')
def public_key_file(tmpdir_factory):
    file = tmpdir_factory.mktemp('keys').join('public_key')
    file.write(pytest.msgapi.public)
    return str(file)


Callback = collections.namedtuple('Callback', ['queue', 'handle_message'])


@pytest.fixture(scope='module')
def callback(event_loop, connection):
    async def create_callback():
        queue = asyncio.Queue()

        async def handle_message(message):
            await queue.put(message)

        return Callback(queue, handle_message)
    return event_loop.run_until_complete(create_callback())


@pytest.fixture(scope='module')
def callback_server_port():
    return unused_tcp_port()


@pytest.fixture(scope='module')
def callback_server(request, event_loop, connection, callback, callback_server_port):
    async def start_callback_server():
        app = e2e.create_application(connection)
        e2e.add_callback_route(connection, app, callback.handle_message)
        runner = web.AppRunner(app)
        await runner.setup()

        cert_path = pytest.msgapi.cert_path
        ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        ssl_context.load_cert_chain(certfile=cert_path)
        site = web.TCPSite(
            runner, host=pytest.msgapi.ip, port=callback_server_port,
            ssl_context=ssl_context)
        await site.start()

        return app, runner, site
    app, runner, site = event_loop.run_until_complete(start_callback_server())

    def fin():
        async def stop_callback_server():
            await site.stop()
            await runner.cleanup()
            await app.cleanup()
        event_loop.run_until_complete(stop_callback_server())

    request.addfinalizer(fin)


@pytest.fixture(scope='module')
def callback_client(request, event_loop, callback_server):
    async def create_client():
        # Note: This is ONLY required because we are using a self-signed certificate
        #       for test purposes.
        connector_ = aiohttp.TCPConnector(ssl=False)
        session_ = aiohttp.ClientSession(connector=connector_)
        return connector_, session_
    connector, session = event_loop.run_until_complete(create_client())

    def fin():
        event_loop.run_until_complete(session.close())

    request.addfinalizer(fin)
    return session


@pytest.fixture(scope='module')
def callback_send(callback_client, callback_server_port, connection):
    async def send(message):
        # Get data from message
        nonce, data = await message.send(get_data_only=True)

        # Create callback parameters
        params = {
            'from': connection.id,
            'to': message.to_id,
            'messageId': hashlib.sha256(message.to_id.encode('ascii')).hexdigest()[16:],
            'date': str(time.time()),
            'nonce': binascii.hexlify(nonce).decode('ascii'),
            'box': binascii.hexlify(data).decode('ascii'),
        }

        # Calculate MAC
        message = ''.join((params['from'], params['to'], params['messageId'],
                           params['date'], params['nonce'], params['box']))
        message = message.encode('ascii')
        encoded_secret = connection.secret.encode('ascii')
        hmac_ = hmac.new(encoded_secret, msg=message, digestmod=hashlib.sha256)
        params['mac'] = hmac_.hexdigest()

        # Send message
        url = 'https://{}:{}/gateway_callback'.format(
            pytest.msgapi.ip, callback_server_port)
        return await callback_client.post(url, data=params)

    return send


@pytest.fixture(scope='module')
def callback_receive(event_loop, callback, callback_server):
    async def receive(timeout=3.0):
        coroutine = asyncio.wait_for(callback.queue.get(), timeout)
        return await coroutine

    return receive



"""
The command line interface for the Threema gateway service.
"""
import binascii
import os
import re

import aiohttp
import click
import logbook
import logbook.more

from threema.gateway import Connection
from threema.gateway import __version__ as _version
from threema.gateway import (
    e2e,
    feature_level,
    simple,
    util,
)
from threema.gateway.key import (
    HMAC,
    Key,
)
from threema.gateway.util import AioRunMixin

_logging_handler = None
_logging_levels = {
    1: logbook.CRITICAL,
    2: logbook.ERROR,
    3: logbook.WARNING,
    4: logbook.NOTICE,
    5: logbook.INFO,
    6: logbook.DEBUG,
    7: logbook.TRACE,
}

# Apply mock URL when starting CLI in debug mode
_test_port = os.environ.get('THREEMA_TEST_API')
if _test_port is not None:
    _mock_url = 'http://{}:{}'.format('127.0.0.1', _test_port)
    Connection.urls = {key: value.replace('https://msgapi.threema.ch', _mock_url)
                       for key, value in Connection.urls.items()}
    click.echo(('WARNING: Currently running in test mode!'
                'The Threema Gateway Server will not be contacted!'), err=True)


class _MockConnection(AioRunMixin):
    def __init__(self, private_key, public_key, identity=None):
        super().__init__(blocking=False)
        self.key = private_key
        self._public_key = public_key
        self.id = identity

    async def get_public_key(self, _):
        return self._public_key


@click.group()
@click.option('-v', '--verbosity', type=click.IntRange(0, len(_logging_levels)),
              default=0, help="Logging verbosity.")
@click.option('-c', '--colored', is_flag=True, help='Colourise logging output.')
@click.option('-vf', '--verify-fingerprint', is_flag=True,
              help='Verify the certificate fingerprint.')
@click.option('--fingerprint', type=str, help='A hex-encoded fingerprint.')
@click.pass_context
def cli(ctx, verbosity, colored, verify_fingerprint, fingerprint):
    """
    Command Line Interface. Use --help for details.
    """
    if verbosity > 0:
        # Enable logging
        util.enable_logging(level=_logging_levels[verbosity])

        # Get handler class
        if colored:
            handler_class = logbook.more.ColorizedStderrHandler
        else:
            handler_class = logbook.StderrHandler

        # Set up logging handler
        handler = handler_class(level=_logging_levels[verbosity])
        handler.push_application()
        global _logging_handler
        _logging_handler = handler

    # Fingerprint
    if fingerprint is not None:
        fingerprint = binascii.unhexlify(fingerprint)

    # Store on context
    ctx.obj = {
        'verify_fingerprint': verify_fingerprint,
        'fingerprint': fingerprint
    }


@cli.command(short_help='Show version information.', help="""
Show the current version of the Python SDK and the implemented feature
level.
""")
def version():
    click.echo('Version: {}'.format(_version))
    click.echo('Feature Level: {}'.format(feature_level))


@cli.command(short_help='Encrypt a text message.', help="""
Encrypt standard input using the given sender PRIVATE KEY and recipient
PUBLIC KEY. Prints two lines to standard output: first the nonce (hex),
and then the encrypted box (hex).
""")
@click.argument('private_key')
@click.argument('public_key')
@util.aio_run
async def encrypt(private_key, public_key):
    # Get key instances
    private_key = util.read_key_or_key_file(private_key, Key.Type.private)
    public_key = util.read_key_or_key_file(public_key, Key.Type.public)

    # Read text from stdin
    text = click.get_text_stream('stdin').read()

    # Print nonce and message as hex
    connection = _MockConnection(private_key, public_key)
    message = e2e.TextMessage(connection, text=text, to_id='')
    nonce, message = await message.send(get_data_only=True)
    click.echo()
    click.echo(binascii.hexlify(nonce))
    click.echo(binascii.hexlify(message))


@cli.command(short_help='Decrypt a text message.', help="""
Decrypt standard input using the given recipient PRIVATE KEY and sender PUBLIC KEY.
The NONCE must be given on the command line, and the box (hex) on standard input.
Prints the decrypted text message to standard output.
""")
@click.argument('private_key')
@click.argument('public_key')
@click.argument('nonce')
@util.aio_run
async def decrypt(private_key, public_key, nonce):
    # Get key instances
    private_key = util.read_key_or_key_file(private_key, Key.Type.private)
    public_key = util.read_key_or_key_file(public_key, Key.Type.public)

    # Convert nonce to bytes
    nonce = binascii.unhexlify(nonce)

    # Read message from stdin and convert to bytes
    message = click.get_text_stream('stdin').read()
    message = binascii.unhexlify(message)

    # Unpack message
    connection = _MockConnection(private_key, public_key)
    parameters = {'from_id': '', 'message_id': '', 'date': ''}
    message = await e2e.Message.receive(connection, parameters, nonce, message)

    # Ensure that this is a text message
    if message.type is not e2e.Message.Type.text_message:
        raise TypeError('Cannot decrypt message type {} in CLI'.format(message.type))

    # Print text
    click.echo()
    click.echo(message.text)


@cli.command(short_help='Generate a new key pair.', help="""
Generate a new key pair and write the PRIVATE and PUBLIC keys to
the respective files.
""")
@click.argument('private_key_file')
@click.argument('public_key_file')
def generate(private_key_file, public_key_file):
    # Generate key pair and hexlify both keys
    private_key, public_key = [Key.encode(key) for key in Key.generate_pair()]

    # Write keys to files
    with open(private_key_file, 'w') as sk_file, open(public_key_file, 'w') as pk_file:
        sk_file.write(private_key + '\n')
        pk_file.write(public_key + '\n')


# noinspection PyShadowingBuiltins
@cli.command(short_help='Hash an email address or phone number.', help="""
Hash an email address or a phone number for identity lookup.
Prints the hash in hex.
""")
@click.option('-e', '--email', help='An email address.')
@click.option('-p', '--phone', help='A phone number in E.164 format.')
def hash(**arguments):
    mode = {key: value for key, value in arguments.items() if value is not None}

    # Check that either email or phone has been specified
    if len(mode) != 1:
        error = 'Please specify exactly one email address or one phone number.'
        raise click.ClickException(error)

    # Unpack message and hash type
    hash_type, message = mode.popitem()

    # Email or phone?
    if hash_type == 'email':
        message = message.lower().strip()
    else:
        message = re.sub(r'[^0-9]', '', message)

    click.echo(HMAC.hash(message, hash_type).hexdigest())


@cli.command(short_help='Derive the public key from the private key.', help="""
Derive the public key that corresponds with the given PRIVATE KEY.
""")
@click.argument('private_key')
def derive(private_key):
    # Get private key instance and derive public key
    private_key = util.read_key_or_key_file(private_key, Key.Type.private)
    public_key = Key.derive_public(private_key)

    # Return hex encoded public key
    click.echo(Key.encode(public_key))


@cli.command(short_help='Send a text message using simple mode.', help="""
Send atext  message from standard input with server-side encryption to the given ID.
FROM is the API identity and SECRET is the API secret.
Prints the message ID on success.
""")
@click.argument('to')
@click.argument('from')
@click.argument('secret')
@click.pass_context
@util.aio_run
async def send_simple(ctx, **arguments):
    # Read message from stdin
    text = click.get_text_stream('stdin').read().strip()

    # Create connection
    connection = Connection(arguments['from'], arguments['secret'], **ctx.obj)
    async with connection:
        # Create message
        message = simple.TextMessage(
            connection=connection,
            to_id=arguments['to'],
            text=text
        )

        # Send message
        click.echo()
        click.echo(await message.send())


@cli.command(short_help='Send a text message using end-to-end mode.', help="""
Encrypt standard input and send the text message to the given ID.
FROM is the API identity and SECRET is the API secret.
Prints the message ID on success.
""")
@click.argument('to')
@click.argument('from')
@click.argument('secret')
@click.argument('private_key')
@click.option('-k', '--public-key', help="""
The public key of the recipient. Will be fetched automatically if not provided.
""")
@click.pass_context
@util.aio_run
async def send_e2e(ctx, **arguments):
    # Get key instances
    private_key = util.read_key_or_key_file(arguments['private_key'], Key.Type.private)
    if arguments['public_key'] is not None:
        public_key = util.read_key_or_key_file(arguments['public_key'], Key.Type.public)
    else:
        public_key = None

    # Read message from stdin
    text = click.get_text_stream('stdin').read().strip()

    # Create connection
    connection = Connection(
        identity=arguments['from'],
        secret=arguments['secret'],
        key=private_key,
        **ctx.obj
    )

    async with connection:
        # Create message
        message = e2e.TextMessage(
            connection=connection,
            to_id=arguments['to'],
            key=public_key,
            text=text
        )

        # Send message
        click.echo()
        click.echo(await message.send())


@cli.command(short_help='Send an image using end-to-end mode.', help="""
Encrypt and send an image ('jpeg' or 'png') to the given ID.
FROM is the API identity and SECRET is the API secret.
IMAGE_PATH is a relative or absolute path to an image.
Prints the message ID on success.
""")
@click.argument('to')
@click.argument('from')
@click.argument('secret')
@click.argument('private_key')
@click.argument('image_path')
@click.option('-k', '--public-key', help="""
The public key of the recipient. Will be fetched automatically if not provided.
""")
@click.pass_context
@util.aio_run
async def send_image(ctx, **arguments):
    # Get key instances
    private_key = util.read_key_or_key_file(arguments['private_key'], Key.Type.private)
    if arguments['public_key'] is not None:
        public_key = util.read_key_or_key_file(arguments['public_key'], Key.Type.public)
    else:
        public_key = None

    # Create connection
    connection = Connection(
        identity=arguments['from'],
        secret=arguments['secret'],
        key=private_key,
        **ctx.obj
    )

    async with connection:
        # Create message
        message = e2e.ImageMessage(
            connection=connection,
            to_id=arguments['to'],
            key=public_key,
            image_path=arguments['image_path']
        )

        # Send message
        click.echo(await message.send())


@cli.command(short_help='Send a video using end-to-end mode.', help="""
Encrypt and send a video ('mp4') including a thumbnail to the given ID.
FROM is the API identity and SECRET is the API secret.
VIDEO_PATH is a relative or absolute path to a video.
THUMBNAIL_PATH is a relative or absolute path to a thumbnail.
Prints the message ID on success.
""")
@click.argument('to')
@click.argument('from')
@click.argument('secret')
@click.argument('private_key')
@click.argument('video_path')
@click.argument('thumbnail_path')
@click.option('-k', '--public-key', help="""
The public key of the recipient. Will be fetched automatically if not provided.
""")
@click.option('-d', '--duration', help="""
Duration of the video in seconds. Defaults to 0.
""", default=0)
@click.pass_context
@util.aio_run
async def send_video(ctx, **arguments):
    # Get key instances
    private_key = util.read_key_or_key_file(arguments['private_key'], Key.Type.private)
    if arguments['public_key'] is not None:
        public_key = util.read_key_or_key_file(arguments['public_key'], Key.Type.public)
    else:
        public_key = None

    # Create connection
    connection = Connection(
        identity=arguments['from'],
        secret=arguments['secret'],
        key=private_key,
        **ctx.obj
    )

    async with connection:
        # Create message
        message = e2e.VideoMessage(
            connection=connection,
            to_id=arguments['to'],
            key=public_key,
            duration=arguments['duration'],
            video_path=arguments['video_path'],
            thumbnail_path=arguments['thumbnail_path']
        )

        # Send message
        click.echo(await message.send())


@cli.command(short_help='Send a file using end-to-end mode.', help="""
Encrypt and send a file to the given ID, optionally with a thumbnail.
FROM is the API identity and SECRET is the API secret.
FILE_PATH is a relative or absolute path to a file.
Prints the message ID on success.
""")
@click.argument('to')
@click.argument('from')
@click.argument('secret')
@click.argument('private_key')
@click.argument('file_path')
@click.option('-k', '--public-key', help="""
The public key of the recipient. Will be fetched automatically if not provided.
""")
@click.option('-t', '--thumbnail-path', help="""
The relative or absolute path to a thumbnail.
""")
@click.pass_context
@util.aio_run
async def send_file(ctx, **arguments):
    # Get key instances
    private_key = util.read_key_or_key_file(arguments['private_key'], Key.Type.private)
    if arguments['public_key'] is not None:
        public_key = util.read_key_or_key_file(arguments['public_key'], Key.Type.public)
    else:
        public_key = None

    # Create connection
    connection = Connection(
        identity=arguments['from'],
        secret=arguments['secret'],
        key=private_key,
        **ctx.obj
    )

    async with connection:
        # Create message
        message = e2e.FileMessage(
            connection=connection,
            to_id=arguments['to'],
            key=public_key,
            file_path=arguments['file_path'],
            thumbnail_path=arguments['thumbnail_path']
        )

        # Send message
        click.echo(await message.send())


@cli.command(short_help='Lookup a Threema ID or the public key.', help="""
Lookup the public key of the Threema ID or the ID linked to either the
given email address or the given phone number.
FROM is the API identity and SECRET is the API secret.
""")
@click.argument('from')
@click.argument('secret')
@click.option('-e', '--email', help='An email address.')
@click.option('-p', '--phone', help='A phone number in E.164 format.')
@click.option('-i', '--id', help='A Threema ID.')
@click.pass_context
@util.aio_run
async def lookup(ctx, **arguments):
    modes = ['email', 'phone', 'id']
    mode = {key: value for key, value in arguments.items()
            if key in modes and value is not None}

    # Check that one of the modes has been selected
    if len(mode) != 1:
        error = 'Please specify exactly one ID, one email address or one phone number.'
        raise click.ClickException(error)

    # Create connection
    connection = Connection(arguments['from'], secret=arguments['secret'], **ctx.obj)
    async with connection:
        # Do lookup
        if 'id' in mode:
            public_key = await connection.get_public_key(arguments['id'])
            click.echo(Key.encode(public_key))
        else:
            click.echo(await connection.get_id(**mode))


@cli.command(short_help='Lookup the reception capabilities of a Threema ID', help="""
Lookup the reception capabilities of a Threema ID.
FROM is the API identity and SECRET is the API secret.
Prints a set of capabilities in alphabetical order on success.
""")
@click.argument('from')
@click.argument('secret')
@click.argument('id')
@click.pass_context
@util.aio_run
async def capabilities(ctx, **arguments):
    # Create connection
    connection = Connection(arguments['from'], arguments['secret'], **ctx.obj)
    async with connection:
        # Lookup and format returned capabilities
        coroutine = connection.get_reception_capabilities(arguments['id'])
        capabilities_ = await coroutine
        click.echo(', '.join(sorted(capability.value for capability in capabilities_)))


# noinspection PyShadowingBuiltins
@cli.command(short_help='Get the number of credits left on the account', help="""
Retrieve the number of credits left on the used account.
FROM is the API identity and SECRET is the API secret.
""")
@click.argument('from')
@click.argument('secret')
@click.pass_context
@util.aio_run
async def credits(ctx, **arguments):
    # Create connection
    connection = Connection(arguments['from'], arguments['secret'], **ctx.obj)
    async with connection:
        # Get and print credits
        click.echo(await connection.get_credits())


def main():
    exc = None
    try:
        cli()
    except aiohttp.client_exceptions.ServerFingerprintMismatch:
        error = 'Fingerprints did not match!'
    except Exception as exc_:
        error = str(exc_)
        exc = exc_
    else:
        error = None

    # Print error (if any)
    if error is not None:
        click.echo('An error occurred:', err=True)
        click.echo(error, err=True)

        # Re-raise
        if exc is not None:
            raise exc

    # Remove logging handler
    if _logging_handler is not None:
        _logging_handler.pop_application()


if __name__ == '__main__':
    main()



"""
The tests provided in this module have been tested for compliance with
the Threema Gateway server. Obviously, the simulated server does not
completely mimic the behaviour of the Gateway server.
"""
import asyncio

import pytest

from threema.gateway import (
    ReceptionCapability,
    e2e,
    simple,
)
from threema.gateway.exception import (
    BlobServerError,
    CreditsServerError,
    IDServerError,
    KeyServerError,
    MessageServerError,
    ReceptionCapabilitiesServerError,
)


async def download_blob(connection, blob_id):
    response = await connection.download(blob_id)
    return await response.read()


async def get_latest_blob_ids(server, connection):
    return await asyncio.gather(*[download_blob(connection, blob_id)
                                for blob_id in server.latest_blob_ids])


class TestLookupPublicKey:
    @pytest.mark.asyncio
    async def test_invalid_identitiy(self, invalid_connection):
        with pytest.raises(KeyServerError) as exc_info:
            await invalid_connection.get_public_key('ECHOECHO')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_invalid_length(self, connection):
        with pytest.raises(KeyServerError) as exc_info:
            await connection.get_public_key('TEST')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_unknown_id(self, connection):
        with pytest.raises(KeyServerError) as exc_info:
            await connection.get_public_key('00000000')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_valid_id(self, connection, server):
        key = await connection.get_public_key('ECHOECHO')
        assert key.hex_pk() == server.echoecho_key

    @pytest.mark.asyncio
    async def test_cache_expiration(self, connection, server):
        connection.get_public_key.cache_clear()
        for _ in range(10):
            await connection.get_public_key('ECHOECHO')
        cache_info = connection.get_public_key.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 9
        await asyncio.sleep(0.2)
        await connection.get_public_key('ECHOECHO')
        cache_info = connection.get_public_key.cache_info()
        # For some reason, the cache is not always being cleared and I don't want
        # Travis to fail all the time
        assert cache_info.misses == 1 or cache_info.misses == 2

    @pytest.mark.asyncio
    async def test_cache_hits(self, connection, server):
        connection.get_public_key.cache_clear()
        for _ in range(1000):
            await self.test_valid_id(connection, server)
        cache_info = connection.get_public_key.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByPhone:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            await invalid_connection.get_id(phone='44123456789')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_unknown_phone(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(phone='44987654321')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_invalid_phone(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(phone='-12537825318')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_valid_phone(self, connection):
        id_ = await connection.get_id(phone='44123456789')
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    async def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            await self.test_valid_phone(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByPhoneHash:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            await invalid_connection.get_id(phone_hash='invalid_hash')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_invalid_length(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af14421'
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(phone_hash=phone_hash)
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    async def test_odd_length(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214'
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(phone_hash=phone_hash)
        # Note: This status code might not be intended and may change in the future
        assert exc_info.value.status == 500

    @pytest.mark.asyncio
    async def test_invalid_phone_hash(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(phone_hash='1234')
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    async def test_unknown_phone_hash(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214a'
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(phone_hash=phone_hash)
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_valid_phone_hash(self, connection):
        phone_hash = '98b05f6eda7a878f6f016bdcdc9db6eb61a6b190e814ff787142115af144214c'
        id_ = await connection.get_id(phone_hash=phone_hash)
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    async def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            await self.test_valid_phone_hash(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByEmail:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            await invalid_connection.get_id(email='echoecho@example.com')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_unknown_email(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(email='somemail@example.com')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_invalid_email(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(email='invalid')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_valid_email(self, connection):
        id_ = await connection.get_id(email='echoecho@example.com')
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    async def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            await self.test_valid_email(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestLookupIDByEmailHash:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(IDServerError) as exc_info:
            await invalid_connection.get_id(email_hash='invalid_hash')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_invalid_length(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669'
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(email_hash=email_hash)
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    async def test_odd_length(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e'
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(email_hash=email_hash)
        # Note: This status code might not be intended and may change in the future
        assert exc_info.value.status == 500

    @pytest.mark.asyncio
    async def test_invalid_email_hash(self, connection):
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(email_hash='1234')
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    async def test_unknown_email_hash(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e1'
        with pytest.raises(IDServerError) as exc_info:
            await connection.get_id(email_hash=email_hash)
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_valid_email_hash(self, connection):
        email_hash = '45a13d422b40f81936a9987245d3f6d9064c90607273af4f578246b4484669e2'
        id_ = await connection.get_id(email_hash=email_hash)
        assert id_ == 'ECHOECHO'

    @pytest.mark.asyncio
    async def test_cache_hits(self, connection):
        connection.get_id.cache_clear()
        for _ in range(1000):
            await self.test_valid_email_hash(connection)
        cache_info = connection.get_id.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestReceptionCapabilities:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(ReceptionCapabilitiesServerError) as exc_info:
            await invalid_connection.get_reception_capabilities('ECHOECHO')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_invalid_length(self, connection):
        with pytest.raises(ReceptionCapabilitiesServerError) as exc_info:
            await connection.get_reception_capabilities('TEST')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_unknown_id(self, connection):
        with pytest.raises(ReceptionCapabilitiesServerError) as exc_info:
            await connection.get_reception_capabilities('00000000')
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_valid_id(self, connection):
        key = await connection.get_reception_capabilities('ECHOECHO')
        assert key == {
            ReceptionCapability.text,
            ReceptionCapability.image,
            ReceptionCapability.video,
            ReceptionCapability.file
        }

    @pytest.mark.asyncio
    async def test_cache_hits(self, connection):
        connection.get_reception_capabilities.cache_clear()
        for _ in range(1000):
            await self.test_valid_id(connection)
        cache_info = connection.get_reception_capabilities.cache_info()
        assert cache_info.misses == 1
        assert cache_info.hits == 999


class TestCredits:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(CreditsServerError) as exc_info:
            await invalid_connection.get_credits()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_valid(self, connection):
        assert (await connection.get_credits()) == 100


class TestUploadBlob:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(BlobServerError) as exc_info:
            await invalid_connection.upload(b'\x01')
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_insufficient_credits(self, nocredit_connection):
        with pytest.raises(BlobServerError) as exc_info:
            await nocredit_connection.upload(b'\x01')
        assert exc_info.value.status == 402

    @pytest.mark.asyncio
    async def test_just_ok(self, connection, server):
        blob_id = await connection.upload(bytes(20 * (2**20)))
        assert len(blob_id) == 32
        # Note: Remove big blob because further tests may hang
        del server.blobs[blob_id]

    @pytest.mark.asyncio
    async def test_too_big(self, connection, server):
        with pytest.raises(BlobServerError) as exc_info:
            blob_id = await connection.upload(bytes((20 * (2**20)) + 1))
            # Note: Remove big blob because further tests may hang
            del server.blobs[blob_id]
        assert exc_info.value.status == 413

    @pytest.mark.asyncio
    async def test_zero(self, connection):
        with pytest.raises(BlobServerError) as exc_info:
            await connection.upload(b'')
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    async def test_file(self, connection):
        assert len(await connection.upload(b'\x01')) == 32


class TestDownloadBlob:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection, blob_id):
        with pytest.raises(BlobServerError) as exc_info:
            await (await invalid_connection.download(blob_id)).read()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_invalid_id(self, connection):
        with pytest.raises(BlobServerError) as exc_info:
            await (await connection.download('f' * 15)).read()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_unknown_id(self, connection):
        with pytest.raises(BlobServerError) as exc_info:
            await (await connection.download('f' * 16)).read()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_file(self, connection, blob_id, blob):
        response = await connection.download(blob_id)
        assert (await response.read()) == blob

    @pytest.mark.asyncio
    async def test_no_credits(self, nocredit_connection, blob_id, blob):
        response = await nocredit_connection.download(blob_id)
        assert (await response.read()) == blob


class TestSendSimple:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection):
        with pytest.raises(MessageServerError) as exc_info:
            await simple.TextMessage(
                connection=invalid_connection,
                to_id='ECHOECHO',
                text='Hello'
            ).send()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_insufficient_credits(self, nocredit_connection):
        with pytest.raises(MessageServerError) as exc_info:
            await simple.TextMessage(
                connection=nocredit_connection,
                to_id='ECHOECHO',
                text='Hello'
            ).send()
        assert exc_info.value.status == 402

    @pytest.mark.asyncio
    async def test_message_too_long(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            await simple.TextMessage(
                connection=connection,
                to_id='ECHOECHO',
                text='0' * 3501
            ).send()
        assert exc_info.value.status == 413

    @pytest.mark.asyncio
    async def test_unknown_id(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            await simple.TextMessage(
                connection=connection,
                to_id='00000000',
                text='Hello'
            ).send()
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    async def test_unknown_email(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            await simple.TextMessage(
                connection=connection,
                email='somemail@example.com',
                text='Hello'
            ).send()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_unknown_phone(self, connection):
        with pytest.raises(MessageServerError) as exc_info:
            await simple.TextMessage(
                connection=connection,
                phone='44987654321',
                text='Hello'
            ).send()
        assert exc_info.value.status == 404

    @pytest.mark.asyncio
    async def test_via_id(self, connection):
        id_ = await simple.TextMessage(
            connection=connection,
            to_id='ECHOECHO',
            text='0' * 3500
        ).send()
        assert id_ == '0' * 16

    @pytest.mark.asyncio
    async def test_via_email(self, connection):
        id_ = await simple.TextMessage(
            connection=connection,
            email='echoecho@example.com',
            text='Hello'
        ).send()
        assert id_ == '0' * 16

    @pytest.mark.asyncio
    async def test_via_phone(self, connection):
        id_ = await simple.TextMessage(
            connection=connection,
            phone='44123456789',
            text='Hello'
        ).send()
        assert id_ == '0' * 16


class TestSendE2E:
    @pytest.mark.asyncio
    async def test_invalid_identity(self, invalid_connection, server):
        with pytest.raises(MessageServerError) as exc_info:
            await e2e.TextMessage(
                connection=invalid_connection,
                to_id='ECHOECHO',
                key=server.echoecho_encoded_key,
                text='Hello'
            ).send()
        assert exc_info.value.status == 401

    @pytest.mark.asyncio
    async def test_insufficient_credits(self, nocredit_connection, server):
        with pytest.raises(MessageServerError) as exc_info:
            await e2e.TextMessage(
                connection=nocredit_connection,
                to_id='ECHOECHO',
                key=server.echoecho_encoded_key,
                text='Hello'
            ).send()
        assert exc_info.value.status == 402

    @pytest.mark.asyncio
    async def test_message_too_long(self, connection, raw_message):
        with pytest.raises(MessageServerError) as exc_info:
            await raw_message(
                connection=connection,
                to_id='ECHOECHO',
                nonce=b'0' * 24,
                message=b'1' * 4001
            ).send()
        assert exc_info.value.status == 413

    @pytest.mark.asyncio
    async def test_unknown_id(self, connection, server):
        with pytest.raises(MessageServerError) as exc_info:
            await e2e.TextMessage(
                connection=connection,
                to_id='00000000',
                key=server.echoecho_encoded_key,
                text='Hello'
            ).send()
        assert exc_info.value.status == 400

    @pytest.mark.asyncio
    async def test_raw(self, connection, raw_message):
        id_ = await raw_message(
            connection=connection,
            to_id='ECHOECHO',
            nonce=b'0' * 24,
            message=b'1' * 4000
        ).send()
        assert id_ == '1' * 16

    @pytest.mark.asyncio
    async def test_via_id(self, connection):
        id_ = await e2e.TextMessage(
            connection=connection,
            to_id='ECHOECHO',
            text='Hello'
        ).send()
        assert id_ == '1' * 16

    @pytest.mark.asyncio
    async def test_via_id_and_key(self, connection, server):
        id_ = await e2e.TextMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            text='Hello'
        ).send()
        assert id_ == '1' * 16

    @pytest.mark.asyncio
    async def test_image(self, connection, server):
        server.latest_blob_ids = []
        id_ = await e2e.ImageMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            image_path=server.threema_jpg
        ).send()
        assert id_ == '1' * 16
        assert len(server.latest_blob_ids) == 1
        assert all((await get_latest_blob_ids(server, connection)))

    @pytest.mark.asyncio
    async def test_video(self, connection, server):
        server.latest_blob_ids = []
        id_ = await e2e.VideoMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            duration=1,
            video_path=server.threema_mp4,
            thumbnail_path=server.threema_jpg,
        ).send()
        assert id_ == '1' * 16
        assert len(server.latest_blob_ids) == 2
        assert all((await get_latest_blob_ids(server, connection)))

    @pytest.mark.asyncio
    async def test_file(self, connection, server):
        server.latest_blob_ids = []
        id_ = await e2e.FileMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            file_path=server.threema_jpg
        ).send()
        assert id_ == '1' * 16
        assert len(server.latest_blob_ids) == 1
        assert all((await get_latest_blob_ids(server, connection)))

    @pytest.mark.asyncio
    async def test_file_with_thumbnail(self, connection, server):
        server.latest_blob_ids = []
        id_ = await e2e.FileMessage(
            connection=connection,
            to_id='ECHOECHO',
            key=server.echoecho_encoded_key,
            file_path=server.threema_jpg,
            thumbnail_path=server.threema_jpg
        ).send()
        assert id_ == '1' * 16
        assert len(server.latest_blob_ids) == 2
        assert all((await get_latest_blob_ids(server, connection)))



"""
This API can be used to send text messages to any Threema user, and to
receive incoming messages and delivery receipts.

There are two main modes of operation:

* Basic mode (server-based encryption)
    - The server handles all encryption for you.
    - The server needs to know the private key associated with your
      Threema API identity.
    - Incoming messages and delivery receipts are not supported.

* End-to-end encrypted mode
    - The server doesn't know your private key.
    - Incoming messages and delivery receipts are supported.
    - You need to run software on your side to encrypt each message
      before it can be sent, and to decrypt any incoming messages or
      delivery receipts.

The mode that you can use depends on the way your account was set up.

.. moduleauthor:: Lennart Grahl <lennart.grahl@gmail.com>
"""
import itertools

from . import _gateway
from . import exception as _exception
from ._gateway import *  # noqa
from .exception import *  # noqa

__author__ = 'Lennart Grahl <lennart.grahl@gmail.com>'
__status__ = 'Production'
__version__ = '4.0.0'
feature_level = 3

__all__ = tuple(itertools.chain(
    ('feature_level',),
    ('bin', 'simple', 'e2e', 'key', 'util'),
    _gateway.__all__,
    _exception.__all__,
))



import libnacl
import pytest

from threema.gateway import (
    e2e,
    key,
)


class TestCrypto:
    def test_incorrect_nonce(self):
        key_pair = key.Key.generate_pair()
        data_in = b'meow'
        nonce = b'0' * 23
        with pytest.raises(ValueError) as exc_info:
            e2e._pk_encrypt(key_pair, data_in, nonce=nonce)
        assert 'Invalid nonce size' in str(exc_info.value)

    def test_incorrect_ciphertext(self):
        key_pair = key.Key.generate_pair()
        data_in = b'meow'
        nonce = b'0' * 24
        with pytest.raises(libnacl.CryptError) as exc_info:
            _, data_encrypted = e2e._pk_encrypt(key_pair, data_in, nonce=nonce)
            e2e._pk_decrypt(key_pair, nonce, data_encrypted + b'0')
        assert 'decrypt' in str(exc_info.value)

    def test_valid(self):
        key_pair = key.Key.generate_pair()
        data_in = b'meow'
        nonce = b'0' * 24
        _, data_encrypted = e2e._pk_encrypt(key_pair, data_in, nonce=nonce)
        data_out = e2e._pk_decrypt(key_pair, nonce, data_encrypted)
        assert data_in == data_out



"""
Provides classes and functions for the end-to-end encryption mode.
"""
import abc
import binascii
import collections
import datetime
import enum
import functools
import hashlib
import hmac
import json
import mimetypes
import os
import struct
from typing import Tuple

import libnacl
import libnacl.encode
import libnacl.public
import libnacl.secret
from aiohttp import web

from . import ReceptionCapability
from .exception import (
    CallbackError,
    DirectionError,
    MessageError,
    MissingCapabilityError,
    UnsupportedMimeTypeError,
)
from .key import Key
from .util import (
    AioRunMixin,
    ViewIOReader,
    ViewIOWriter,
    aio_run_proxy,
    randint,
)

__all__ = (
    'BLOB_ID_LENGTH',
    'MAX_HTTP_REQUEST_SIZE',
    'CallbackContext',
    'handle_callback',
    'create_application',
    'add_callback_route',
    'Message',
    'DeliveryReceipt',
    'TextMessage',
    'ImageMessage',
    'VideoMessage',
    'FileMessage',
)

BLOB_ID_LENGTH = 16

# A box can contain up to 4000 bytes, so this should be sufficient.
# The remaining POST parameters aren't that big.
# See: https://gateway.threema.ch/en/developer/api
MAX_HTTP_REQUEST_SIZE = 8192


def _pk_encrypt(key_pair: Tuple[Key, Key], data: bytes, nonce: bytes = None):
    """
    Encrypt data by using public-key encryption.

    Arguments:
        - `key_pair`: A tuple containing our private key and the public
          key of the recipient.
        - `data`: Raw data.
        - `nonce`: A predefined nonce.

    Raises `libnacl.CryptError` in case the data could not be encrypted.
    Raises `ValueError` in other cases.

    Return a tuple of bytes containing the nonce and the encrypted
    data.
    """
    # Assemble and encrypt the payload
    private, public = key_pair
    box = libnacl.public.Box(sk=private, pk=public)
    return box.encrypt(data, nonce=nonce, pack_nonce=False)


def _pk_decrypt(key_pair: Tuple[Key, Key], nonce: bytes, data: bytes):
    """
    Decrypt data by using public-key decryption.

    Arguments:
        - `key_pair`: A tuple containing our private key and the public
          key of the sender.
        - `nonce`: The nonce of the encrypted data.
        - `data`: Encrypted data.

    Raises `libnacl.CryptError` in case the data could not be decrypted.
    Raises `ValueError` in other cases.

    Return the decrypted data.
    """
    # Decrypt payload
    private, public = key_pair
    box = libnacl.public.Box(sk=private, pk=public)
    return box.decrypt(data, nonce=nonce)


def _sk_encrypt(key, data, nonce=None):
    """
    Encrypt data by using secret-key encryption.

    Arguments:
        - `key`: A secret key.
        - `data`: Data (bytes).
        - `nonce`: A predefined nonce.

    Return a tuple of bytes containing the nonce and the encrypted
    data.
    """
    # Assemble and encrypt the payload
    box = libnacl.secret.SecretBox(key=key)
    # Note: Workaround for libnacl which lacks `pack_nonce` option
    # (see: https://github.com/saltstack/libnacl/pull/61)
    # return box.encrypt(data, nonce=nonce, pack_nonce=False)
    data = box.encrypt(data, nonce=nonce)
    nonce_length = libnacl.crypto_secretbox_NONCEBYTES
    return data[:nonce_length], data[nonce_length:]


def _sk_decrypt(key, nonce, data):
    """
    Decrypt data by using secret-key decryption.

    Arguments:
        - `key`: A secret key.
        - `nonce`: The nonce of the encrypted message.
        - `data`: Encrypted data (bytes).

    Return the decrypted data.
    """
    # Decrypt payload
    box = libnacl.secret.SecretBox(key=key)
    return box.decrypt(data, nonce=nonce)


def _validate_hmac(encoded_secret, expected_mac, response):
    try:
        message = ''.join((
            response['from'],
            response['to'],
            response['messageId'],
            response['date'],
            response['nonce'],
            response['box'],
        )).encode('ascii')
    except UnicodeError as exc:
        raise CallbackError(400, 'Cannot concatenate HMAC message') from exc
    hmac_ = hmac.new(encoded_secret, msg=message, digestmod=hashlib.sha256)
    actual_mac = hmac_.hexdigest()
    if not hmac.compare_digest(expected_mac, actual_mac):
        raise CallbackError(400, 'MACs do not match')


CallbackContext = collections.namedtuple('CallbackContext', [
    'encoded_secret', 'connection', 'message_handler', 'receive_handler',
])


async def handle_callback(context, request):
    try:
        response = await request.post()

        # Unpack fields
        try:
            from_id = response['from']
            to_id = response['to']
            message_id = response['messageId']
            date = response['date']
            nonce = response['nonce']
            data = response['box']
            mac = response['mac']
        except KeyError as exc:
            raise CallbackError(400, 'Could not unpack required fields') from exc

        # Validate HMAC and ID
        _validate_hmac(context.encoded_secret, mac, response)
        if to_id != context.connection.id:
            raise CallbackError(400, 'IDs do not match')

        # Validate from id length
        if len(from_id) != 8:
            raise CallbackError(400, "Invalid 'from' value")

        # Convert date and message id
        try:
            message_id = binascii.unhexlify(message_id)
        except binascii.Error as exc:
            raise CallbackError(400, 'Invalid message ID') from exc
        try:
            date = datetime.datetime.fromtimestamp(float(date))
        except (ValueError, TypeError) as exc:
            raise CallbackError(400, 'Invalid date') from exc

        # Convert nonce and data
        try:
            nonce = binascii.unhexlify(nonce)
            data = binascii.unhexlify(data)
        except binascii.Error as exc:
            raise CallbackError(400, 'Invalid nonce or data') from exc

        # Unpack message
        try:
            message = await context.receive_handler(context.connection, {
                'from_id': from_id,
                'message_id': message_id,
                'date': date,
            }, nonce, data)
        except MessageError as exc:
            raise CallbackError(400, str(exc)) from exc

        # Pass message to handler
        await context.message_handler(message)

        # Respond with 'OK'
        return web.Response(status=200)
    except CallbackError as exc:
        # Note: For security reasons, we do not send the reason
        return web.Response(status=exc.status)
    except Exception:
        raise


def create_application(connection):
    application = web.Application(client_max_size=MAX_HTTP_REQUEST_SIZE)
    return application


def add_callback_route(
    connection, application, message_handler,
    path='/gateway_callback', receive_handler=None,
):
    if receive_handler is None:
        receive_handler = Message.receive
    context = CallbackContext(
        connection.secret.encode('ascii'),
        connection,
        message_handler,
        receive_handler,
    )
    application.router.add_routes([web.post(
        path, functools.partial(handle_callback, context))])


# TODO: Update docstring (arguments)
class Message(AioRunMixin, metaclass=abc.ABCMeta):
    """
    A message class all end-to-end mode messages are derived from.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Attributes:
        - `connection`: An instance of a connection.
        - `type_`: The message type.
        - `id_`: Threema ID of the sender (incoming) or recipient
          (outgoing).
        - `key`: The public key of the sender/recipient. Will be
          fetched from the server if not supplied.
        - `key_file`: A file where the private key is stored in. Can
          be used instead of passing the key directly.
    """
    async_functions = {
        'key',
        'send',
        'receive',
        'pack',
        'unpack',
        'check_capabilities',
        'get_encrypt_key_pair',
        'get_decrypt_key_pair',
        'encrypt',
    }
    nonce = {
        'video': (b'\x00' * 23) + b'\x01',
        'file': (b'\x00' * 23) + b'\x01',
        'thumbnail': (b'\x00' * 23) + b'\x02'
    }

    @enum.unique
    class Type(enum.IntEnum):
        """
        The type of a message.
        """
        text_message = 0x01
        image_message = 0x02
        video_message = 0x13
        file_message = 0x17
        delivery_receipt = 0x80

    @enum.unique
    class Direction(enum.IntEnum):
        """
        Incoming or outgoing message.
        """
        outgoing = 1
        incoming = 2

    @classmethod
    def get_message_class(cls, type_):
        """
        Return the corresponding :class:`Message` class for a
        :class:`Message.Type`.

        Arguments:
            - `type_`: A :class:`Message.Type`.

        Raises :exc:`KeyError` if no matching message class could be
        found.
        """
        if getattr(cls, '_message_classes', None) is None:
            cls._message_classes = {
                cls.Type.text_message: TextMessage,
                cls.Type.image_message: ImageMessage,
                cls.Type.video_message: VideoMessage,
                cls.Type.file_message: FileMessage,
                cls.Type.delivery_receipt: DeliveryReceipt,
            }
        return cls._message_classes[type_]

    def __init__(
            self, connection, type_,
            key=None, key_file=None,
            to_id=None, from_data=None
    ):
        super().__init__(blocking=connection.blocking)
        connection = connection.unwrap

        # Get direction
        if from_data is not None:
            direction = self.Direction.incoming
        else:
            direction = self.Direction.outgoing

        # Check required parameters
        if direction == self.Direction.outgoing:
            if to_id is None:
                message = "Parameter 'to_id' is required for outgoing messages."
                raise ValueError(message)
            from_id = connection.id
            message_id = date = None
        elif direction == self.Direction.incoming:
            keys = ('from_id', 'message_id', 'date')
            if any((key not in from_data for key in keys)):
                message = 'Parameters {} are required for incoming messages.'
                raise ValueError(message.format(keys))
            from_id = from_data['from_id']
            to_id = connection.id
            message_id = from_data['message_id']
            date = from_data['date']
        else:
            raise ValueError('Invalid direction value')

        # Required for both directions
        self._connection = connection
        self._direction = direction
        self._type = type_
        self._key = None
        self._key_file = None
        self.key = key
        self.key_file = key_file

        # Values depending on direction
        self.to_id = to_id  # Always set
        self.from_id = from_id  # Always set
        self.message_id = message_id  # None if outgoing
        self.date = date  # None if outgoing

    @property
    def type(self):
        """Return the :class:`Message.Type`."""
        return self._type

    # TODO: Raises?
    @property
    async def key(self):
        """
        Get the public key of the recipient (outgoing messages) or the
        sender (incoming messages). Will be request from the server
        if necessary. Note that the getter is a coroutine!

        Set the public key of the recipient or sender. The key will be
        decoded if required.
        """
        if self._key is None:
            self._key = await self._connection.get_public_key(self.to_id)
        return self._key

    @key.setter
    def key(self, key):
        if isinstance(key, str):
            key = Key.decode(key, Key.Type.public)
        self._key = key

    @property
    def key_file(self):
        """
        Get the path of the recipients public key file.

        Set the public key of the recipient by reading it from a file.
        """
        return self._key_file

    @key_file.setter
    def key_file(self, key_file):
        if key_file is not None:
            with open(key_file) as file:
                self.key = file.readline().strip()
        self._key_file = key_file

    # TODO: Raises?
    async def send(self, get_data_only=False):
        """
        Send a message.

        Raises :exc:`DirectionError` in case the message is not intended
        to be sent.
        """
        if self._direction != self.Direction.outgoing:
            raise DirectionError('Parameters are missing to send the message')
        writer = ViewIOWriter()

        # Pack type
        try:
            writer.writeexactly(struct.pack('<B', self.type.value))
        except struct.error as exc:
            raise MessageError('Could not pack type') from exc

        # Get content data
        await self.pack(writer)

        # Generate 0 < padding < 256
        padding_length = randint(1, 255)
        # Add padding to data
        writer.writeexactly(bytes([padding_length] * padding_length))

        # Encrypt message
        nonce, data = await self.encrypt(writer.getvalue())
        if get_data_only:
            return nonce, data

        # Send message
        return await self._connection.send_e2e(**{
            'to': self.to_id,
            'nonce': binascii.hexlify(nonce).decode('ascii'),
            'box': binascii.hexlify(data).decode('ascii')
        })

    # TODO: Raises?
    @classmethod
    async def receive(cls, connection, parameters, nonce, data):
        """
        Return a :class:`Message` instance from an encrypted message.

        The argument `parameters` contains the following items:
            - `from_id`: Sender's identity.
            - `message_id`: Message ID assigned by the sender as bytes.
            - `date`: A :class:´datetime.datetime` instance.

        Arguments:
            - `connection`: A :class:`Connection` instance.
            - `parameters`: A :class:`dict` containing parameters
              (see above).
            - `nonce`: A :class:`bytes`-like instance containing the
              messages' nonce.
            - `data`: A :class:`bytes`-like instance containing the
              encrypted message.

        Raises:
            - :exc:`MessageError` in case the message is invalid.
        """
        # Decrypt message
        key_pair = await cls.get_decrypt_key_pair(connection, parameters)
        data = cls.decrypt(nonce, data, key_pair)

        # Unpack type and padding length
        try:
            type_, *_ = struct.unpack('<B', data[:1])
            padding_length, *_ = struct.unpack('<B', data[-1:])
        except struct.error as exc:
            raise MessageError('Could not unpack type and padding') from exc

        # Validate type and get message class
        try:
            type_ = cls.Type(type_)
            class_ = cls.get_message_class(type_)
        except ValueError:
            raise MessageError('Cannot handle type: {}'.format(type_))

        # Remove type and padding from data
        reader = ViewIOReader(data[1:-padding_length])

        # Unpack message
        return await class_.unpack(connection, parameters, key_pair, reader)

    @abc.abstractmethod
    async def pack(self, writer):
        """
        Pack payload data.

        Arguments:
            - `writer`: A :class:`ViewIOWriter` instance to write data
              into.

        Raises `exc`:MessageError` in case the message could not be
        packed.
        """
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    async def unpack(cls, connection, parameters, key_pair, reader):
        """
        Return a :class:`Message` instance from raw payload data.

        The argument `parameters` contains the following items:
            - `from_id`: Sender's identity.
            - `message_id`: Message ID assigned by the sender as bytes.
            - `date`: A :class:´datetime.datetime` instance.

        Arguments:
            - `connection`: A :class:`Connection` instance.
            - `parameters`: A :class:`dict` containing parameters
              (see above).
            - `key_pair`: A tuple containing our private key and the
              public key of the sender.
            - `reader`: A :class:`ViewIOReader` containing the payload.

        Raises:
            - :exc:`MessageError` in case the message is invalid.
        """
        raise NotImplementedError

    # TODO: Raises?
    async def check_capabilities(self, required_capabilities):
        """
        Test for capabilities of a recipient.

        Arguments:
            - `required_capabilities`: A set of capabilities that are
              required.

        Raise :class:`MissingCapabilityError` in case that one or more
        capabilities are missing.
        """
        # Check capabilities of a recipient
        capabilities_coroutine = self._connection.get_reception_capabilities(self.to_id)
        recipient_capabilities = await capabilities_coroutine
        if not required_capabilities <= recipient_capabilities:
            missing_capabilities = required_capabilities - recipient_capabilities
            raise MissingCapabilityError(missing_capabilities)

    # TODO: Raises?
    @property
    async def get_encrypt_key_pair(self):
        """
        Return a tuple containing our private key and the public key of
        the recipient.
        """
        private = self._connection.key
        public = await self.key
        return private, public

    # TODO: Raises?
    @classmethod
    async def get_decrypt_key_pair(cls, connection, parameters):
        """
        Return a tuple containing our private key and the public key of
        the sender.

        Arguments:
            - `connection`: A :class:`Connection` instance.
            - `parameters`: Parameters passed to
              :method:`Message.unpack`.
        """
        private = connection.key
        public = await connection.get_public_key(parameters['from_id'])
        return private, public

    async def encrypt(self, data, key_pair=None, nonce=None):
        """
        Encrypt data.

        Arguments:
            - `data`: Raw data (bytes).
            - `key_pair`: A tuple containing our private key and the
              public key of the recipient.
            - A predefined nonce.

        Raises :exc:`MessageError` in case data could not be encrypted.

        Return a tuple of bytes containing the nonce and the encrypted
        data.
        """
        # Key pair specified?
        if key_pair is None:
            key_pair = await self.get_encrypt_key_pair

        # Encrypt
        try:
            return _pk_encrypt(key_pair, data, nonce=nonce)
        except (ValueError, libnacl.CryptError) as exc:
            raise MessageError('Could not encrypt data') from exc

    @classmethod
    def decrypt(cls, nonce, data, key_pair):
        """
        Decrypt data.

        Arguments:
            - `nonce`: The nonce of the encrypted data.
            - `data`: Encrypted data (bytes).
            - `key_pair`: A tuple containing our private key and the
              public key of the sender.

        Raises :exc:`MessageError` in case data could not be decrypted.

        Return decrypted data as :class:´bytes`.
        """
        # Decrypt
        try:
            return _pk_decrypt(key_pair, nonce, data)
        except (ValueError, libnacl.CryptError) as exc:
            raise MessageError('Could not decrypt data') from exc


# TODO: Update docstring (arguments)
@aio_run_proxy
class DeliveryReceipt(Message):
    """
    A delivery receipt that can be received in end-to-end mode.
    Each delivery receipt message confirms the receipt of one
    or multiple regular text messages.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Arguments:
        - `payload`: The remaining byte sequence of the message.
    """
    async_functions = {
        'pack',
        'unpack',
    }

    @enum.unique
    class ReceiptType(enum.IntEnum):
        """
        Describes message receipt types.
        """
        received = 0x01
        read = 0x02
        user_acknowledge = 0x03
        user_decline = 0x04

    def __init__(
            self, connection, receipt_type=None, message_ids=None,
            from_data=None, **kwargs
    ):
        super().__init__(connection, Message.Type.delivery_receipt,
                         from_data=from_data, **kwargs)
        if self._direction == self.Direction.outgoing:
            if receipt_type is None or message_ids is None:
                message = "Parameters 'receipt_type' and 'message_ids' are required"
                raise ValueError(message)
            self.receipt_type = receipt_type
            self.message_ids = message_ids
        else:
            self.receipt_type = from_data.get('receipt_type')
            self.message_ids = from_data.get('message_ids')

    def __str__(self):
        ids = (binascii.hexlify(id_).decode('ascii') for id_ in self.message_ids)
        return 'Delivery receipt({}): {}'.format(self.receipt_type.name, ', '.join(ids))

    async def pack(self, writer):
        # Pack content
        try:
            writer.writeexactly(struct.pack('<B', self.receipt_type.value))
        except struct.error as exc:
            raise MessageError('Could not pack receipt type') from exc
        for message_id in self.message_ids:
            writer.writeexactly(message_id)

    @classmethod
    async def unpack(cls, connection, parameters, key_pair, reader):
        # Check length
        length = len(reader)
        if length < 9 or (length - 1) % 8 != 0:
            raise MessageError('Invalid length')

        # Unpack content
        formatter = '<B' + '8s' * int((length - 1) // 8)
        data = reader.readexactly(length)
        try:
            receipt_type, *message_ids = struct.unpack(formatter, data)
        except struct.error as exc:
            message = 'Could not unpack receipt type and message ids'
            raise MessageError(message) from exc

        # Validate receipt type
        try:
            receipt_type = cls.ReceiptType(receipt_type)
        except ValueError as exc:
            raise MessageError('Unknown receipt type: {}'.format(receipt_type)) from exc

        # Return instance
        return cls(connection, from_data=dict(parameters, **{
            'receipt_type': receipt_type,
            'message_ids': message_ids,
        }))


# TODO: Update docstring (arguments)
@aio_run_proxy
class TextMessage(Message):
    """
    A text message.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Arguments for a new message:
        - `connection`: An instance of a connection.
        - `id`: Threema ID of the recipient.
        - `key`: The public key of the recipient. Will be fetched from
           the server if not supplied.
        - `key_file`: A file where the private key is stored in. Can
          be used instead of passing the key directly.
        - `text`: Message text. Will be encrypted automatically.

    Arguments for an existing message:
        - `payload`: The remaining byte sequence of a decrypted
          message.
    """
    async_functions = {
        'pack',
        'unpack',
    }

    def __init__(self, connection, text=None, from_data=None, **kwargs):
        super().__init__(connection, Message.Type.text_message,
                         from_data=from_data, **kwargs)
        if self._direction == self.Direction.outgoing:
            if text is None:
                raise ValueError("Parameter 'text' required")
            self.text = text
        else:
            self.text = from_data.get('text')

    def __str__(self):
        return self.text

    async def pack(self, writer):
        # Encode text
        try:
            text = self.text.encode('utf-8')
        except UnicodeError as exc:
            raise MessageError('Could not encode text') from exc

        # Add text
        writer.writeexactly(text)

    @classmethod
    async def unpack(cls, connection, parameters, key_pair, reader):
        # Get text
        text = bytes(reader.readexactly(len(reader)))

        # Decode text
        try:
            text = text.decode('utf-8')
        except UnicodeError as exc:
            raise MessageError('Could not decode text') from exc

        # Return instance
        return cls(connection, from_data=dict(parameters, **{
            'text': text,
        }))


# TODO: Update docstring (arguments)
@aio_run_proxy
class ImageMessage(Message):
    """
    An image message.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Arguments for a new message:
        - `connection`: An instance of a connection.
        - `id`: Threema ID of the recipient.
        - `key`: The public key of the recipient. Will be fetched from
           the server if not supplied.
        - `key_file`: A file where the private key is stored in. Can
          be used instead of passing the key directly.
        - `image_path`: A file where the image is stored in.

    Arguments for an existing message:
        - `payload`: The remaining byte sequence of a decrypted
          message.
    """
    async_functions = {
        'pack',
        'unpack',
    }
    allowed_mime_types = {
        'image/jpg',
        'image/jpeg',
        'image/png'
    }
    required_capabilities = {
        ReceptionCapability.image
    }
    _formatter = '<{}sI{}s'.format(BLOB_ID_LENGTH, libnacl.crypto_box_NONCEBYTES)

    def __init__(
            self, connection,
            image=None, mime_type=None, image_path=None,
            from_data=None, **kwargs
    ):
        super().__init__(connection, Message.Type.image_message,
                         from_data=from_data, **kwargs)
        if self._direction == self.Direction.outgoing:
            image_and_mime = all((param is not None for param in [image, mime_type]))
            path_param = image_path is not None
            if sum((1 for param in (image_and_mime, path_param) if param)) != 1:
                raise ValueError(("Either 'image' and 'mime_type' or 'image_path' "
                                  "need to be specified"))
            self._image = image
            self._mime_type = mime_type
            self._image_path = image_path
        else:
            self._image = from_data['image']
            self._mime_type = from_data['mime_type']
            self._image_path = None

    @property
    def image(self):
        """
        Return the image as :class:`bytes`.

        Raises :exc:`OSError` in case the image could not be read from
        the specified path.
        """
        self._read_image()
        return self._image

    @property
    def mime_type(self):
        """
        Return the mime type as :class:`str`.
        """
        self._read_image()
        return self._mime_type

    def _read_image(self):
        """
        Read and store the image as :class:`bytes`.

        Raises :exc:`OSError` in case the image could not be read from
        the specified path.
        """
        if self._image is None:
            with open(self._image_path, mode='rb') as file:
                # Read content
                self._image = file.read()

            # Guess the mime type
            mime_type, _ = mimetypes.guess_type(self._image_path)
            if mime_type not in self.allowed_mime_types:
                raise UnsupportedMimeTypeError(mime_type)
            self._mime_type = mime_type

    async def pack(self, writer):
        # Check capabilities of recipient
        await self.check_capabilities(self.required_capabilities)

        # Encrypt and upload image
        image_nonce, image_data = await self.encrypt(self.image)
        blob_id = await self._connection.upload(image_data)
        try:
            blob_id = binascii.unhexlify(blob_id)
        except binascii.Error as exc:
            raise MessageError('Could not convert hex-encoded blob id') from exc

        # Pack blob id, image length and image nonce
        try:
            data = struct.pack(self._formatter, blob_id, len(self.image), image_nonce)
        except struct.error as exc:
            raise MessageError('Could not pack blob id, length and nonce') from exc

        # Add data
        writer.writeexactly(data)

    @classmethod
    async def unpack(cls, connection, parameters, key_pair, reader):
        # Unpack blob id, image length and image nonce
        length = struct.calcsize(cls._formatter)
        try:
            data = struct.unpack(cls._formatter, reader.readexactly(length))
        except struct.error as exc:
            raise MessageError('Could not unpack blob id, length and nonce') from exc
        blob_id, image_length, image_nonce = data

        # Download and decrypt image
        blob_id = binascii.hexlify(blob_id).decode('ascii')
        response = await connection.download(blob_id)
        image_data = await response.read()
        image = cls.decrypt(image_nonce, image_data, key_pair)

        # Return instance
        return cls(connection, from_data=dict(parameters, **{
            'image': image,
            'mime_type': None,  # TODO: Guess mime type from bytes
        }))


# TODO: Update docstring (arguments)
@aio_run_proxy
class VideoMessage(Message):
    """
    A video message including a thumbnail.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Arguments for a new message:
        - `connection`: An instance of a connection.
        - `id`: Threema ID of the recipient.
        - `key`: The public key of the recipient. Will be fetched from
           the server if not supplied.
        - `key_file`: A file where the private key is stored in. Can
          be used instead of passing the key directly.
        - `duration`: The duration of the video in seconds.
        - `video_path`: A file where the video is stored in.
        - `thumbnail_path`: The path to a thumbnail of the file.

    Arguments for an existing message:
        - `payload`: The remaining byte sequence of a decrypted
          message.
    """
    async_functions = {
        'pack',
        'unpack',
    }
    required_capabilities = {
        ReceptionCapability.video
    }
    _formatter = '<H{}sI{}sI{}s'.format(
        BLOB_ID_LENGTH, BLOB_ID_LENGTH, libnacl.crypto_box_SECRETKEYBYTES)

    def __init__(
            self, connection,
            duration=0,
            video=None, video_path=None,
            thumbnail_content=None, thumbnail_path=None,
            from_data=None, **kwargs
    ):
        super().__init__(connection, Message.Type.video_message,
                         from_data=from_data, **kwargs)
        if self._direction == self.Direction.outgoing:
            path_param = video_path is not None
            if sum((1 for param in (video, path_param) if param)) != 1:
                raise ValueError(("Either 'video' or 'video_path' need to be specified"))
            if sum((1 for param in (thumbnail_content, thumbnail_path) if param)) != 1:
                raise ValueError(("Either 'thumbnail_content' or 'thumbnail_path' "
                                  "need to be specified"))
            self.duration = duration
            self._video = video
            self._video_path = video_path
            self._thumbnail_content = thumbnail_content
            self._thumbnail_path = thumbnail_path
        else:
            self.duration = from_data['duration']
            self._video = from_data['video']
            self._video_path = None
            self._thumbnail_content = from_data['thumbnail_content']
            self._thumbnail_path = None

    @property
    def video(self):
        """
        Return the video as :class:`bytes`.

        Raises :exc:`OSError` in case the video could not be read from
        the specified path.
        """
        self._read_video()
        return self._video

    @property
    def thumbnail_content(self):
        """
        Return the thumbnails' content as :class:`bytes`.

        Raises :exc:`OSError` in case the thumbnails' content could
        not be read from the specified path.
        """
        self._read_thumbnail()
        return self._thumbnail_content

    def _read_video(self):
        """
        Read and store the video as :class:`bytes`.

        Raises :exc:`OSError` in case the video could not be read from
        the specified path.
        """
        if self._video is None:
            with open(self._video_path, mode='rb') as file:
                # Read content
                self._video = file.read()

    def _read_thumbnail(self):
        """
        Read and store the thumbnails' content as :class:`bytes`.

        Raises :exc:`OSError` in case the thumbnails' content could
        not be read from the specified path.
        """
        if self._thumbnail_content is None and self._thumbnail_path is not None:
            with open(self._thumbnail_path, mode='rb') as file:
                # Read content
                self._thumbnail_content = file.read()

    async def pack(self, writer):
        # Check capabilities of recipient
        await self.check_capabilities(self.required_capabilities)

        # Generate a symmetric key for the video and its thumbnail
        key, _ = Key.generate_secret_key()

        # Encrypt and upload video by a newly generated symmetric key
        _, video_data = _sk_encrypt(key, self.video, nonce=self.nonce['video'])
        video_id = await self._connection.upload(video_data)
        try:
            video_id = binascii.unhexlify(video_id)
        except binascii.Error as exc:
            raise MessageError('Could not convert hex-encoded blob id') from exc

        # Encrypt and upload thumbnail
        _, thumbnail_data = _sk_encrypt(
            key, self.thumbnail_content, nonce=self.nonce['thumbnail'])
        thumbnail_id = await self._connection.upload(thumbnail_data)
        try:
            thumbnail_id = binascii.unhexlify(thumbnail_id)
        except binascii.Error as exc:
            raise MessageError('Could not convert hex-encoded blob id') from exc

        # Pack duration, blob ids, video/thumbnail length and the secret key
        try:
            data = struct.pack(
                self._formatter,
                self.duration,
                video_id,
                len(self.video),
                thumbnail_id,
                len(self.thumbnail_content),
                key,
            )
        except struct.error as exc:
            raise MessageError(
                'Could not pack duration, blob ids, length and key') from exc

        # Add data
        writer.writeexactly(data)

    @classmethod
    async def unpack(cls, connection, parameters, key_pair, reader):
        # Unpack duration, blob ids, video/thumbnail length and the secret key
        length = struct.calcsize(cls._formatter)
        try:
            data = struct.unpack(cls._formatter, reader.readexactly(length))
        except struct.error as exc:
            raise MessageError(
                'Could not unpack duration, blob ids, length and key') from exc
        duration, video_id, video_length, thumbnail_id, thumbnail_length, key = data

        # Download and decrypt thumbnail
        thumbnail_id = binascii.hexlify(thumbnail_id).decode('ascii')
        response = await connection.download(thumbnail_id)
        thumbnail_data = await response.read()
        thumbnail_content = _sk_decrypt(key, cls.nonce['thumbnail'], thumbnail_data)

        # Validate thumbnail content length
        length = len(thumbnail_content)
        if length != thumbnail_length:
            message = 'Thumbnail content length does not match (expected: {}, got: {})'
            raise MessageError(message.format(thumbnail_length, length))

        # Download and decrypt video
        video_id = binascii.hexlify(video_id).decode('ascii')
        response = await connection.download(video_id)
        video_data = await response.read()
        video = _sk_decrypt(key, cls.nonce['file'], video_data)

        # Validate video content length
        length = len(video)
        if length != video_length:
            message = 'Video content length does not match (expected: {}, got: {})'
            raise MessageError(message.format(video_length, length))

        # Return instance
        return cls(connection, from_data=dict(parameters, **{
            'duration': duration,
            'video': video,
            'thumbnail_content': thumbnail_content,
        }))


# TODO: Update docstring (arguments)
@aio_run_proxy
class FileMessage(Message):
    """
    A file message including a thumbnail.

    If the connection passed to the constructor is in blocking mode, then all
    methods on this class will be blocking too.

    Arguments for a new message:
        - `connection`: An instance of a connection.
        - `id`: Threema ID of the recipient.
        - `key`: The public key of the recipient. Will be fetched from
           the server if not supplied.
        - `key_file`: A file where the private key is stored in. Can
          be used instead of passing the key directly.
        - `file_path`: The path to a file.
        - `thumbnail_path`: The path to a thumbnail of the file.

    Arguments for an existing message:
        - `payload`: The remaining byte sequence of a decrypted
          message.
    """
    async_functions = {
        'pack',
        'unpack',
    }
    required_capabilities = {
        ReceptionCapability.file
    }

    def __init__(
            self, connection,
            file_content=None, mime_type=None, file_name='file', file_path=None,
            thumbnail_content=None, thumbnail_path=None,
            from_data=None, **kwargs
    ):
        super().__init__(connection, Message.Type.file_message,
                         from_data=from_data, **kwargs)
        if self._direction == self.Direction.outgoing:
            file_and_mime = all((param is not None for param
                                 in [file_content, mime_type]))
            path_param = file_path is not None
            if sum((1 for param in (file_and_mime, path_param) if param)) != 1:
                raise ValueError(("Either 'file_content' and 'mime_type' or 'file_path' "
                                  "need to be specified"))
            if sum((1 for param in (thumbnail_content, thumbnail_path) if param)) > 1:
                raise ValueError(("Either 'thumbnail_content' or 'thumbnail_path' may "
                                  "to be specified"))
            self._file_content = file_content
            self._mime_type = mime_type
            self._file_name = file_name
            self._file_path = file_path
            self._thumbnail_content = thumbnail_content
            self._thumbnail_path = thumbnail_path
        else:
            self._file_content = from_data['file_content']
            self._mime_type = from_data['mime_type']
            self._file_name = from_data['file_name']
            self._file_path = None
            self._thumbnail_content = from_data['thumbnail_content']
            self._thumbnail_path = None

    @property
    def file_content(self):
        """
        Return the files' content as :class:`bytes`.

        Raises :exc:`OSError` in case the files' content could not be
        read from the specified path.
        """
        self._read_file()
        return self._file_content

    @property
    def mime_type(self):
        """
        Return the mime type of the file as :class:`str`.
        """
        self._read_file()
        return self._mime_type

    @property
    def thumbnail_content(self):
        """
        Return the thumbnails' content as :class:`bytes`.

        Raises :exc:`OSError` in case the thumbnails' content could
        not be read from the specified path.
        """
        self._read_thumbnail()
        return self._thumbnail_content

    def _read_file(self):
        """
        Read and store the files' content as :class:`bytes`.

        Raises :exc:`OSError` in case the files' content could not be
        read from the specified path.
        """
        if self._file_content is None:
            with open(self._file_path, mode='rb') as file:
                # Read content
                self._file_content = file.read()

            # Guess the mime type
            mime_type, _ = mimetypes.guess_type(self._file_path)
            if mime_type is None:
                # Fallback mime type
                mime_type = 'application/octet-stream'
            self._mime_type = mime_type

    def _read_thumbnail(self):
        """
        Read and store the thumbnails' content as :class:`bytes`.

        Raises :exc:`OSError` in case the thumbnails' content could
        not be read from the specified path.
        """
        if self._thumbnail_content is None and self._thumbnail_path is not None:
            with open(self._thumbnail_path, mode='rb') as file:
                # Read content
                self._thumbnail_content = file.read()

    async def pack(self, writer):
        # Check capabilities of recipient
        await self.check_capabilities(self.required_capabilities)

        # Encrypt and upload file by a newly generated symmetric key
        key, hex_key = Key.generate_secret_key()
        _, file_data = _sk_encrypt(key, self.file_content, nonce=self.nonce['file'])
        file_id = await self._connection.upload(file_data)

        # Build JSON
        if self._file_path is not None:
            file_name = os.path.basename(self._file_path)
        else:
            file_name = self._file_name
        content = {
            'b': file_id,
            'k': hex_key.decode('ascii'),
            'm': self.mime_type,
            'n': file_name,
            's': len(self.file_content),
            'i': 0,
        }

        # Encrypt and upload thumbnail (if any)
        thumbnail_content = self.thumbnail_content
        if thumbnail_content is not None:
            _, thumbnail_data = _sk_encrypt(
                key, thumbnail_content, nonce=self.nonce['thumbnail'])
            thumbnail_id = await self._connection.upload(thumbnail_data)
            # Update JSON
            content['t'] = thumbnail_id

        # Pack payload (compact JSON encoding)
        try:
            content = json.dumps(content, separators=(',', ':')).encode('ascii')
        except UnicodeError as exc:
            raise MessageError('Could not encode JSON') from exc

        # Add payload
        writer.writeexactly(content)

    @classmethod
    async def unpack(cls, connection, parameters, key_pair, reader):
        # Get payload
        try:
            content = bytes(reader.readexactly(len(reader))).decode('ascii')
        except UnicodeError as exc:
            raise MessageError('Could not decode JSON') from exc

        # Unpack payload from JSON
        try:
            content = json.loads(content)
        except UnicodeError as exc:
            raise MessageError('Could not decode JSON') from exc
        except ValueError as exc:
            raise MessageError('Could not load JSON') from exc

        # Unpack JSON
        try:
            file_id = content['b']
            key = binascii.unhexlify(content['k'])
            mime_type = content['m']
            file_name = content['n']
            file_content_length = content['s']
        except KeyError as exc:
            raise MessageError('Invalid JSON payload') from exc
        except binascii.Error as exc:
            raise MessageError('Could not convert hex-encoded secret key') from exc
        thumbnail_id = content.get('t')

        # Download and decrypt thumbnail (if any)
        if thumbnail_id is not None:
            response = await connection.download(thumbnail_id)
            thumbnail_data = await response.read()
            thumbnail_content = _sk_decrypt(key, cls.nonce['thumbnail'], thumbnail_data)
        else:
            thumbnail_content = None

        # Download and decrypt file
        response = await connection.download(file_id)
        file_data = await response.read()
        file_content = _sk_decrypt(key, cls.nonce['file'], file_data)

        # Validate file content length
        length = len(file_content)
        if length != file_content_length:
            message = 'File content length does not match (expected: {}, got: {})'
            raise MessageError(message.format(file_content_length, length))

        # Return instance
        return cls(connection, from_data=dict(parameters, **{
            'file_content': file_content,
            'mime_type': mime_type,
            'file_name': file_name,
            'thumbnail_content': thumbnail_content,
        }))



"""
Contains all exceptions used for the Threema gateway service.
"""
from typing import Dict  # noqa

__all__ = (
    'GatewayError',
    'CallbackError',
    'GatewayServerError',
    'IDError',
    'IDServerError',
    'GatewayKeyError',
    'KeyServerError',
    'ReceptionCapabilitiesError',
    'ReceptionCapabilitiesServerError',
    'CreditsServerError',
    'DirectionError',
    'MessageError',
    'UnsupportedMimeTypeError',
    'MissingCapabilityError',
    'MessageServerError',
    'BlobError',
    'BlobServerError',
)


class GatewayError(Exception):
    """
    General error of this module. All other exceptions are derived from
    this class.
    """


class CallbackError(GatewayError):
    """
    Indicates that the callback does not accept the message. The
    *status* code will be returned to the sender.

    Arguments:
        - `status`: An HTTP status code.
        - `reason`: Reason of the *status*.
    """
    def __init__(self, status, reason):
        self.status = status
        self.reason = reason

    def __str__(self):
        return '[{}] {}'.format(self.status, self.reason)


class GatewayServerError(GatewayError):
    """
    The server has responded with an error code. All other server
    exceptions are derived from this class.

    .. versionchanged:: 1.1.9
       Now only contains an HTTP status code instead of holding whole
       request objects which may not be released.
       Does now return the status code when printed.

    Arguments:
        - `status`: An HTTP status code.
    """
    status_description = {}  # type: Dict[int, str]

    def __init__(self, status):
        self.status = status

    def __str__(self):
        status = self.status

        # Return description for status code
        try:
            return '[{}] {}'.format(status, self.status_description[status])
        except KeyError:
            return 'Unknown error, status code: {}'.format(status)


class IDError(GatewayError):
    """
    A problem before fetching a Threema ID occurred.
    """


class IDServerError(IDError, GatewayServerError):
    """
    The server has responded with an error code while looking up a
    Threema ID.
    """
    status_description = {
        400: 'Supplied hash invalid',
        401: 'API identity or secret incorrect',
        404: 'No matching Threema ID could be found',
        500: 'Temporary internal server error occurred'
    }


class GatewayKeyError(GatewayError):
    """
    A problem with a key occurred.

    .. versionchanged:: 1.1.6
       Previous versions shadowed the builtin exception
       :class:`KeyError`.
    """


class KeyServerError(GatewayKeyError, GatewayServerError):
    """
    The server has responded with an error code while fetching a
    public key.
    """
    status_description = {
        401: 'API identity or secret incorrect',
        404: 'No matching Threema ID could be found',
        500: 'Temporary internal server error occurred'
    }


class ReceptionCapabilitiesError(GatewayError):
    """
    An invalid reception capability has been returned.
    """


class ReceptionCapabilitiesServerError(
    ReceptionCapabilitiesError, GatewayServerError
):
    """
    The server responded with an error code while fetching the reception
    capabilities of a Threema ID.
    """
    status_description = {
        401: 'API identity or secret incorrect',
        404: 'No matching Threema ID could be found',
        500: 'Temporary internal server error occurred'
    }


class CreditsServerError(GatewayServerError):
    """
    The server has responded with an error code while fetching the
    remaining credits.
    """
    status_description = {
        401: 'API identity or secret incorrect',
        500: 'Temporary internal server error occurred'
    }


class DirectionError(GatewayError):
    """
    Indicates that a message can not be processed for the specified
    direction because required parameters are missing.
    """


class MessageError(GatewayError):
    """
    Indicates that a message is invalid.
    """


class UnsupportedMimeTypeError(MessageError):
    """
    Indicates that the supplied file or binary of a message has an
    unsupported mime type.
    """
    def __init__(self, mime_type):
        self.mime_type = mime_type

    def __str__(self):
        return 'Unsupported mime type: {}'.format(self.mime_type)


class MissingCapabilityError(MessageError):
    """
    A capability is missing to send a specific message type.
    """
    def __init__(self, missing_capabilities):
        self.missing_capabilities = missing_capabilities

    def __str__(self):
        return 'Missing capabilities: {}'.format(self.missing_capabilities)


class MessageServerError(MessageError, GatewayServerError):
    """
    The server has responded with an error code while sending a
    message.
    """
    status_description = {
        400: 'Recipient identity is invalid or the account is not setup for the '
             'requested mode',
        401: 'API identity or secret incorrect',
        402: 'Insufficient credits',
        404: 'Phone or email address could not be resolved to a Threema ID',
        413: 'Message too long',
        500: 'Temporary internal server error occurred'
    }


class BlobError(GatewayError):
    """
    Indicates that a blob is invalid. The server has not been contacted,
    yet.
    """


class BlobServerError(BlobError, GatewayServerError):
    """
    The server has responded with an error code while uploading or
    downloading a blob.
    """
    status_description = {
        400: 'Required parameters missing or blob is empty',
        401: 'API identity or secret incorrect',
        402: 'Insufficient credits',
        404: 'No matching blob found for the supplied ID',
        413: 'Blob too big',
        500: 'Temporary internal server error occurred'
    }



import ssl

import logbook
import logbook.more
from aiohttp import web

from threema.gateway import (
    Connection,
    util,
)
from threema.gateway.e2e import (
    add_callback_route,
    create_application,
)


async def handle_message(message):
    print('Got message ({}): {}'.format(repr(message), message))


def serve():
    # Create connection instance
    connection = Connection(
        identity='*YOUR_GATEWAY_THREEMA_ID',
        secret='YOUR_GATEWAY_THREEMA_ID_SECRET',
        key='private:YOUR_PRIVATE_KEY'
    )

    # Create the application and register the handler for incoming messages
    application = create_application(connection)
    add_callback_route(connection, application, handle_message, path='/gateway_callback')

    # Create an SSL context to terminate TLS.
    # Note: It is usually advisable to use a reverse proxy instead in front of
    #       the server that terminates TLS, e.g. Nginx.
    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(certfile='YOUR_CERTFILE', keyfile='YOUR_KEYFILE')

    # Run a server that listens on any interface via port 8443. It will
    # gracefully shut down when Ctrl+C has been pressed.
    web.run_app(application, port=8443, ssl_context=ssl_context)


if __name__ == '__main__':
    util.enable_logging(logbook.WARNING)
    log_handler = logbook.more.ColorizedStderrHandler()
    with log_handler.applicationbound():
        serve()



import ast
import os
import sys

from setuptools import (
    find_packages,
    setup,
)


def get_version():
    path = os.path.join(os.path.dirname(__file__), 'threema', 'gateway', '__init__.py')
    with open(path) as file:
        for line in file:
            if line.startswith('__version__'):
                _, value = line.split('=', maxsplit=1)
                return ast.literal_eval(value.strip())
        else:
            raise Exception('Version not found in {}'.format(path))


def read(file):
    return open(os.path.join(os.path.dirname(__file__), file)).read().strip()


# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

# Import long description
long_description = '\n\n'.join((read('README.rst'), read('CHANGELOG.rst')))

# Check python version
py_version = sys.version_info[:3]
if py_version < (3, 6, 1):
    raise Exception("threema.gateway requires Python >= 3.6.1")

# Test requirements
# Note: These are just tools that aren't required, so a version range
#       is not necessary here.
tests_require = [
    'pytest>=3.1.3,<4',
    'pytest-asyncio>=0.6.0,<0.10',
    'pytest-cov>=2.5.1,<3',
    'flake8==3.7.9',
    'isort==4.3.21',
    'collective.checkdocs>=0.2',
    'Pygments>=2.2.0',  # required by checkdocs
    'mypy==0.800',
]

setup(
    name='threema.gateway',
    version=get_version(),
    packages=find_packages(),
    namespace_packages=['threema'],
    install_requires=[
        'memoization==0.2.3',  # we're using private APIs
        'logbook>=1.1.0,<2',
        'libnacl>=1.5.2,<2',
        'click>=6.7,<7',  # doesn't seem to follow semantic versioning
        'aiohttp>=3.7.3,<4',
        'wrapt>=1.10.10,<2',
    ],
    tests_require=tests_require,
    extras_require={
        ':python_version<="3.4"': [
            'asyncio==3.4.3',
            'pytest-asyncio==0.5.0'
        ],
        ':python_version<="3.5"': [
            'typing>=3.6.1,<3.7',
        ],
        'dev': tests_require,
        'uvloop': ['uvloop>=0.8.0,<2'],
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'threema-gateway = threema.gateway.bin.gateway_client:main',
        ],
    },

    # PyPI metadata
    author='Lennart Grahl',
    author_email='lennart.grahl@gmail.com',
    description=('An API for the Threema gateway service to send and receive '
                 'messages including text, images, files and delivery reports.'),
    long_description=long_description,
    license='MIT License',
    keywords='threema gateway service sdk api',
    url='https://gateway.threema.ch/',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Communications :: Chat',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Security',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Logging',
    ],
)



import pytest

from threema.gateway import e2e


class TestCallback:
    @pytest.mark.asyncio
    async def test_invalid_message(self, connection, callback_send, raw_message):
        outgoing = raw_message(
            connection=connection,
            to_id=pytest.msgapi.id,
            nonce=b'0' * 24,
            message=b'1' * 200
        )
        response = await callback_send(outgoing)
        assert response.status == 400
        await response.release()

    @pytest.mark.asyncio
    async def test_delivery_receipt(self, connection, callback_send, callback_receive):
        outgoing = e2e.DeliveryReceipt(
            connection=connection,
            to_id=pytest.msgapi.id,
            receipt_type=e2e.DeliveryReceipt.ReceiptType.read,
            message_ids=[b'0' * 8, b'1' * 8],
        )
        response = await callback_send(outgoing)
        await response.release()
        incoming = await callback_receive()
        assert outgoing.from_id == incoming.from_id
        assert outgoing.to_id == incoming.to_id
        assert outgoing.receipt_type == incoming.receipt_type
        assert outgoing.message_ids == incoming.message_ids

    @pytest.mark.asyncio
    async def test_text_message(self, connection, callback_send, callback_receive):
        outgoing = e2e.TextMessage(
            connection,
            to_id=pytest.msgapi.id,
            text='私はガラスを食べられます。それは私を傷つけません。!',
        )
        response = await callback_send(outgoing)
        await response.release()
        incoming = await callback_receive()
        assert outgoing.from_id == incoming.from_id
        assert outgoing.to_id == incoming.to_id
        assert outgoing.text == incoming.text

    @pytest.mark.asyncio
    async def test_image_message(
        self, connection, callback_send, callback_receive, server
    ):
        outgoing = e2e.ImageMessage(
            connection,
            to_id=pytest.msgapi.id,
            image_path=server.threema_jpg,
        )
        response = await callback_send(outgoing)
        await response.release()
        incoming = await callback_receive()
        assert outgoing.from_id == incoming.from_id
        assert outgoing.to_id == incoming.to_id
        assert outgoing.image == incoming.image

    @pytest.mark.asyncio
    async def test_video(self, connection, callback_send, callback_receive, server):
        outgoing = e2e.VideoMessage(
            connection,
            to_id=pytest.msgapi.id,
            duration=1,
            video_path=server.threema_mp4,
            thumbnail_path=server.threema_jpg,
        )
        response = await callback_send(outgoing)
        await response.release()
        incoming = await callback_receive()
        assert outgoing.from_id == incoming.from_id
        assert outgoing.to_id == incoming.to_id
        assert outgoing.duration == incoming.duration
        assert outgoing.video == incoming.video
        assert outgoing.thumbnail_content == incoming.thumbnail_content

    @pytest.mark.asyncio
    async def test_file_message(
        self, connection, callback_send, callback_receive, server
    ):
        outgoing = e2e.FileMessage(
            connection,
            to_id=pytest.msgapi.id,
            file_path=server.threema_jpg,
        )
        response = await callback_send(outgoing)
        await response.release()
        incoming = await callback_receive()
        assert outgoing.from_id == incoming.from_id
        assert outgoing.to_id == incoming.to_id
        assert outgoing.file_content == incoming.file_content

    @pytest.mark.asyncio
    async def test_file_message_thumb(
            self, connection, callback_send, callback_receive, server
    ):
        outgoing = e2e.FileMessage(
            connection,
            to_id=pytest.msgapi.id,
            file_path=server.threema_jpg,
            thumbnail_path=server.threema_jpg,
        )
        response = await callback_send(outgoing)
        await response.release()
        incoming = await callback_receive()
        assert outgoing.from_id == incoming.from_id
        assert outgoing.to_id == incoming.to_id
        assert outgoing.file_content == incoming.file_content
        assert outgoing.thumbnail_content == incoming.thumbnail_content
